import sys, os, inspect, imp
import encodings.idna
from os.path import join, exists, isdir, basename
from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location
from seneca.engine.interpreter import SenecaInterpreter, Seneca


class SenecaFinder(MetaPathFinder):

    def find_spec(self, fullname, path, target=None):
        if path is None or path == "":
            path = [os.getcwd()] # top level import --
        if "." in fullname:
            *parents, name = fullname.split(".")
        else:
            name = fullname
        for entry in path:
            if isdir(join(entry, name)):
                # this module has child modules
                filename = join(entry, name, "__init__.py")
                if not exists(filename): open(filename, "w+")
                submodule_locations = [join(entry, name)]
            else:
                filename = join(entry, name)
                if exists(filename+'.py'):
                    submodule_locations = [filename]
                    filename += '.py'
                elif exists(filename+'.sen.py'):
                    filename += '.sen.py'
                    submodule_locations = None
                else:
                    continue

            return spec_from_file_location(fullname, filename, loader=SenecaLoader(filename),
                                           submodule_search_locations=submodule_locations)

        return None # we don't know how to import this

class SenecaLoader(Loader):

    def __init__(self, filename):
        self.filename = filename
        self.tree = None
        self.module_name = basename(filename).split('.')[0]
        with open(self.filename) as f:
            code_str = f.read()
            if 'seneca/libs' in self.filename:
                self.code_obj = compile(code_str, filename=self.filename, mode="exec")
            else:
                self.tree, self.postvalidated, self.prevalidated = Seneca.interface.parse_ast(code_str)
                self.code_obj = compile(self.tree, filename=self.filename, mode="exec")

    def exec_module(self, module):
        Seneca.interface.execute(
            self.code_obj, vars(module), is_main=False
        )
        return module

class RedisFinder:

    def find_module(self, fullname, path=None):
        if fullname.startswith('seneca.contracts'):
            return RedisLoader()
        return None


class RedisLoader:

    def load_module(self, fullname):
        self.module_name = module_name = fullname.split('.')[-1]
        code = Seneca.interface.get_code_obj(module_name)
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = "<%s>" % self.__class__.__name__
        mod.__loader__ = self
        mod.__path__ = []
        mod.__package__ = fullname

        Seneca.interface.execute(
            code, mod.__dict__, is_main=False
        )
        return mod



import time, asyncio, ujson as json, redis, traceback
from seneca.libs.logger import get_logger
from seneca.engine.interface import SenecaInterface
from seneca.constants.config import *
from seneca.engine.conflict_resolution import CRContext
from seneca.engine.book_keeper import BookKeeper
from seneca.engine.util import module_path_for_contract
from collections import deque, defaultdict
from typing import Callable, List
import traceback


SUCC_FLAG = 'SUCC'


class Macros:
    # TODO we need to make sure these keys dont conflict with user stuff in the common layer. I.e. users cannot be
    # creating keys named '_execution' or '_conflict_resolution'
    EXECUTION = '_execution_phase'
    CONFLICT_RESOLUTION = '_conflict_resolution_phase'
    RESET = "_reset_phase"

    ALL_MACROS = [EXECUTION, CONFLICT_RESOLUTION]


class Phase:
    EXEC_TIMEOUT = 14  # Number of seconds client will wait for other clients to finish execution phase
    CR_TIMEOUT = 14  # Number of seconds client will wait for other clients to finish conflict resolution phase
    BLOCK_TIMEOUT = 30  # Number of seconds a pending db will wait until it is at the top (first element) of pending_dbs
    AVAIL_DB_TIMEOUT = 60  # How long the client will wait for a DB to become available when executing a SB
    POLL_INTERVAL = 0.5  # Poll for Phase changes every POLL_INTERVAL seconds

    @staticmethod
    def incr(db, key) -> int:
        return db.incr(key)

    @staticmethod
    def get(db, key) -> int:
        if not db.exists(key):
            return 0
        return int(db.get(key).decode())


class SenecaClient(SenecaInterface):
    def __init__(self, sbb_idx, num_sbb, concurrent_mode=True, loop=None):
        # TODO do we even need to bother with the concurrent_mode flag? We are treating that its always true --davis
        super().__init__()

        name = self.__class__.__name__ + "[sbb-{}]".format(sbb_idx)
        self.log = get_logger(name)
        self.loop = loop or asyncio.get_event_loop()

        self.port = get_redis_port()
        self.password = get_redis_password()

        self.sbb_idx = sbb_idx
        self.num_sb_builders = num_sbb
        self.concurrent_mode = concurrent_mode
        self.max_number_workers = NUM_CACHES

        self.master_db = None  # A redis.StrictRedis instance
        self.active_db = None  # Set to a CRContext instance
        self.available_dbs = deque()  # List of CRContext instances
        self.pending_dbs = deque()  # List of CRContext instances

        # pending_futures tracks active db's that are waiting for other SBBs to finish CR/execution
        # it is a map of input hashes -> {'fut': asyncio.Future, 'data': CRContext, ...}
        self.pending_futures = {}

        # queued_futures tracks execute_sb calls that cannot run immediately because there are no available dbs
        # it is a deque of dicts, containing keys 'input_hash': str, and 'fut': asyncio.Future
        self.queued_futures = deque()

        # DEBUG -- TODO DELETE
        # self.log.important2("\n\n\n this is using the latest code \n\n\n")
        # self.log.important("num caches {}".format(NUM_CACHES))
        # END DEBUG

        self._setup_dbs()

    def _setup_dbs(self):
        self.master_db = redis.StrictRedis(host='localhost', port=self.port, db=MASTER_DB, password=self.password)
        for db_num in range(self.max_number_workers):
            db_client = redis.StrictRedis(host='localhost', port=self.port, db=db_num+DB_OFFSET, password=self.password)
            cr_data = CRContext(working_db=db_client, master_db=self.master_db, sbb_idx=self.sbb_idx)
            self.available_dbs.append(cr_data)

    def flush_all(self):
        """ Flushes all pending/active/available dbs. This effectively 'resets' all databases except master."""
        if self.active_db:
            self.active_db.reset_db()
            self.active_db.reset_run_data()
            self.active_db = None
        for s in (self.pending_dbs, self.available_dbs):
            for cr in s:
                cr.reset_db()
                cr.reset_run_data()
            s.clear()

        for input_hash in self.pending_futures:
            self.pending_futures[input_hash]['fut'].cancel()
            if self.pending_futures[input_hash]['merge_fut']:
                self.pending_futures[input_hash]['merge_fut'].cancel()
        self.pending_futures.clear()

        for data in self.queued_futures:
            data['fut'].cancel()
        self.queued_futures.clear()

        # does it leak some dbs?
        self._setup_dbs()

    def _update_master_db(self, expected_data: CRContext=None):
        assert len(self.pending_dbs) > 0, "No pending dbs to update to master!"

        cr_data = self.pending_dbs.popleft()
        if expected_data:
            assert expected_data == cr_data, "Updated master db with expected cr data with input hash {}, but leftpop " \
                                             "cr data has input hash {}!".format(expected_data.input_hash, cr_data.input_hash)
        input_hash = cr_data.input_hash
        assert input_hash is not None, "Input hash is None! Dev error this should not happen"
        assert cr_data.merged_to_common, "CRData not merged to common yet!"

        self.log.notice("Updating master db for CR data {}".format(cr_data))

        futs = self.pending_futures.pop(input_hash, None)

        if self.sbb_idx == 0:
            assert Phase.get(cr_data.working_db, Macros.EXECUTION) == self.num_sb_builders, \
                "Execution stage incomplete!"
            assert Phase.get(cr_data.working_db, Macros.CONFLICT_RESOLUTION) == self.num_sb_builders, \
                "Conflict resolution stage incomplete!"

            self.log.important("Merging common layer to master db (input_hash={})".format(cr_data.input_hash))
            CRContext.merge_to_master(working_db=cr_data.working_db, master_db=self.master_db)

        self.log.debugv("Resetting run data for input hash {}".format(cr_data.input_hash))
        cr_data.reset_run_data()

        # Only hard reset_db (meaning flush redis data) if all other SBBs have finished updating to master db
        reset_var = Phase.incr(cr_data.working_db, Macros.RESET)  # this will do an atomic incr and get
        if reset_var == self.num_sb_builders:
            self.log.debug("RESET phase finished. Resetting db for input hash {}".format(input_hash))
            cr_data.reset_db()

        cr_data.assert_reset()  # Dev check. Make sure cr_data has been properly reset_db

        self.available_dbs.append(cr_data)

    def update_master_db(self):
        # If no pending dbs to sync to master db, return
        # NOTE: For dev, we raise a proper exception. This should not happen. --davis
        if len(self.pending_dbs) == 0:
            # raise Exception("Attempted to update_master_db, but there are no pending_dbs")
            self.log.warning("Attempted to update_master_db, but there are no pending_dbs")
            return

        cr_data = self.pending_dbs[0]
        input_hash = cr_data.input_hash
        cr_finished = Phase.get(cr_data.working_db, Macros.CONFLICT_RESOLUTION) == self.num_sb_builders
        self.log.info("Attempting to merge to master db for input hash {}".format(input_hash))

        assert input_hash in self.pending_futures, "Input hash {} not in pending_futures {}".format(input_hash, self.pending_futures)
        assert cr_data == self.pending_futures[input_hash]['data'], "something has gone horribly wrong"

        if not cr_finished:
            self.log.info(("Deferring merge for {}".format(cr_data)))
            if self.pending_futures[input_hash]['complete']:
                self.log.debugv("Future already complete! Adding new future to update master db when ready")
                fut = self._ensure_future(self._wait_to_update_master_db(cr_data))
                self.pending_futures[input_hash]['merge_fut'] = fut
            else:
                self.log.debugv("CRContext future not yet complete. Setting merge flag to true.")
                self.pending_futures[input_hash]['merge'] = True
        else:
            self.log.info("Merging immediately for {}".format(cr_data))
            self._update_master_db(expected_data=cr_data)

    def skip_current_db(self):
        if len(self.pending_dbs) == 0:
            self.log.warning("skip_current_db called, but there are no pending dbs! Returning.")
            return

        cr_data = self.pending_dbs.popleft()
        cr_dict = self.pending_futures.pop(cr_data.input_hash)

        assert cr_dict['merge'] is False, "tried to skip current db, but merge flag set to true for CR {} with pending" \
                                          " futures dict {}".format(cr_data, cr_dict)
        cr_dict['fut'].cancel()

    def submit_contract(self, contract):
        self.publish_code_str(contract.contract_name, contract.sender, contract.code, scope={
            'rt': {
                'author': contract.sender,
                'sender': contract.sender,
                'contract': contract.contract_name
            }
        })

    def run_contract(self, contract):
        assert self.active_db, "active_db must be set to run a contract. Did you call _start_sb?"
        assert self.active_db.input_hash, "Input hash not set...davis u done goofed again"

        data = self.active_db
        contract_idx = data.next_contract_idx

        result = self._run_contract(contract, contract_idx=contract_idx, data=data)
        self.active_db.add_contract_result(contract, result)

        # DEBUG
        kwargs = 'contract_name=' + contract.contract_name if hasattr(contract, 'contract_code') else contract.kwargs
        self.log.spam("[EXEC PHASE #{}] exec from sender {} with kwargs {} resulted in state {}"
                      .format(contract_idx, contract.sender, kwargs, data.get_state_for_idx(contract_idx)))
        # END DEBUG

    def _run_contract(self, contract, contract_idx: int, data: CRContext) -> str:
        """ Runs the contract object, and retuns a string representing the result (succ/fail).
        Note: Assumes the BookKeeping info has already been set. """
        # Development sanity checks. data passed in should either be the active_db (if we are in execution phase)
        # or it should be next in line in pending_dbs if we are in CR phase
        assert data is self.active_db or data is self.pending_dbs[0], \
            "Data {} is not active db {} or next pending db {}".format(data, self.active_db, self.pending_dbs[0])

        try:
            # Super sketch hack to differentiate between ContractTransactions and PublishTransactions
            BookKeeper.set_cr_info(sbb_idx=self.sbb_idx, contract_idx=contract_idx, data=data, rt={
                'contract': contract.contract_name
            })
            # TODO not this pls
            if hasattr(contract, 'contract_code'):

                author = contract.sender
                self.publish_code_str(fullname=contract.contract_name, author=author,
                                      code_str=contract.contract_code)
            else:
                mod_path = module_path_for_contract(contract)
                run_info = self.execute_function(module_path=mod_path, sender=contract.sender,
                                                 stamps=contract.stamps_supplied, **contract.kwargs)
                # The following is just for debug info
                stamps_sup = contract.stamps_supplied if contract.stamps_supplied is not None else 0
                stamps_spent = stamps_sup - run_info['remaining_stamps']
                self.log.spam("Running contract from sender {} used {} stamps and returned run_info: {}"
                              .format(contract.sender, stamps_spent, run_info))

            result = SUCC_FLAG

        except Exception as e:
            # TODO -- change this log level for production, as we will get spammed like nuts when contracts fail
            self.log.warning("Contract failed with error:\n{} \ncontract obj: {}".format(traceback.format_exc(), contract))
            result = 'FAIL' + ' -- ' + str(e)
            data.rollback_contract(contract_idx)

        return result

    def _rerun_contracts_for_cr_data(self, cr_data: CRContext):
        """ Reruns any contracts in cr_data, if necessary. This should be done before we merge cr_data to common. """
        # This cr_data should be next in line in pending_dbs if we are rerunning contracts
        assert self.pending_dbs[0] is cr_data, "cr_data {} is not first in line in pending_dbs! First is {}"\
                                               .format(cr_data, self.pending_dbs[0])

        self.log.notice("Rerunning any necessary contracts for CRData with input hash {}".format(cr_data.input_hash))
        for i in cr_data.iter_rerun_indexes():
            result = self._run_contract(cr_data.contracts[i], contract_idx=i, data=cr_data)
            cr_data.update_contract_result(contract_idx=i, result=result)

    def _can_start_next_sb(self) -> bool:
        """
        Returns True if we this client can start the next subblock, and False otherwise. A client can start a sb if:
        - There is no active db set
        - There is at least 1 available_db
        - All other SBBs are finished with that first available_db (meaning it has been hard reset_db and merged to master)
        """
        return (not self.active_db and len(self.available_dbs) > 0) and \
               (Phase.get(self.available_dbs[0].working_db, Macros.RESET) == 0)

    def execute_sb(self, input_hash: str, contracts: list, completion_handler: Callable[[CRContext], None]):
        """
        Begins execution of a sub-block with the given list of contracts and the input hash. Once execution
        and conflict resolution is finished, completion_handler is called with a CRContext instance as the only arg.

        If there are no available db's, AND the we have exceeded the allowed nubmer of queued db's, this will return
        False. Otherwise, if a DB is immediately available, or if there is room in the queue, this method will return
        True.
        """
        assert input_hash not in self.pending_futures, "SB with input hash {} is already pending!".format(input_hash)
        for data in self.queued_futures:
            if data['input_hash'] == input_hash:
                raise Exception("Input hash {} already in queued futures {}".format(input_hash, self.queued_futures))
        assert input_hash is not None, "Input hash cannot be None!"

        if self._can_start_next_sb():
            self._execute_sb(input_hash, contracts, completion_handler)

        else:
            if len(self.queued_futures) >= MAX_SB_QUEUE_SIZE:
                self.log.warning("Maximum number ({}) of queueud sub-blocks has been reached! Not scheduling sub-block "
                                 "execution for input hash {}. All queued futures: {}"
                                 .format(MAX_SB_QUEUE_SIZE, input_hash, self.queued_futures))
                return False

            self.log.info("No available dbs. Queueing up future to execute sb for input hash {}".format(input_hash))
            fut = self._ensure_future(self._wait_and_execute_sb(input_hash, contracts, completion_handler))
            self.queued_futures.append({'input_hash': input_hash, 'fut': fut})

        return True

    def _execute_sb(self, input_hash: str, contracts: list, completion_handler: Callable[[CRContext], None]):
        self.log.info(">>> Executing sub block for input hash {} >>>".format(input_hash))

        self._start_sb(input_hash)
        for c in contracts:
            self.run_contract(c)
        self._end_sb(completion_handler)

    def _start_sb(self, input_hash: str):
        assert self.active_db is None, "Attempted to _start_sb, but active_db is already set! Did you end the " \
                                       "previous subblock with _end_sb?"
        assert self._can_start_next_sb(), "Attempted to start a new sub block, but cannot start next sb!!! Dev error!"

        self.active_db = self.available_dbs.popleft()
        self.active_db.assert_reset()  # Dev checks, make sure the CRContext has been properly reset_db
        self.active_db.input_hash = input_hash
        self.log.important("Starting sb with CRData {}".format(self.active_db))

    def _end_sb(self, completion_handler: Callable[[CRContext], None]):
        """
        Ends the current sub block. Specifically, this method:
         - Increments the EXECUTION phase variable
         - Pushes active_db onto the top of the pending_db stack
         - Ensures the _wait_and_merge_to_common future (see doc string on that func for more details)
         - Once the _wait_and_merge_to_common future is done, the callback is trigger
        """
        assert self.active_db, "Active db not set! Did you call _start_sb?"
        self.log.info("<<< Ending sub block for input hash {} <<<".format(self.active_db.input_hash))

        Phase.incr(self.active_db.working_db, Macros.EXECUTION)

        future = self._ensure_future(self._wait_and_merge_to_common(self.active_db, completion_handler))
        self.pending_futures[self.active_db.input_hash] = {'fut': future, 'data': self.active_db, 'merge': False,
                                                           'complete': False, 'merge_fut': None}

        self.pending_dbs.append(self.active_db)
        self.active_db = None  # we really don't care, but might be useful initially for error checking

    async def _wait_and_execute_sb(self, input_hash: str, contracts: list, completion_handler: Callable[[CRContext], None]):
        await self._wait_for_available_db(input_hash)

        assert self.queued_futures[0]['input_hash'] == input_hash, "Input hash {} does not match next queued hash {}" \
            .format(input_hash, self.queued_futures[0]['input_hash'])
        self.queued_futures.popleft()

        self._execute_sb(input_hash, contracts, completion_handler)

    async def _wait_and_merge_to_common(self, cr_data: CRContext, completion_handler: Callable[[CRContext], None]):
        """
        - Waits for all other SBBs to finish execution. Raises an error if this takes longer than Phase.EXEC_TIMEOUT
        - Waits for this cr_data to be first in line in 'pending_dbs' (bottom of stack)
        - Waits for all subblocks before it finish conflict resolution. Raises an error if it takes longer than Phase.CR_TIMEOUT
        - Starts conflict resolution, which involves:
            - Rerun any necessary contracts
            - Merges cr_data to common layer, and then increment the CONFLICT_RESOLUTION phase variable
        - If the 'merge' flag is set by an earlier update_master_db call, then this is done next
        """
        await self._wait_for_execution_stage(cr_data)

        if self.pending_dbs[0] is not cr_data:
            await self._wait_until_top_of_pending(cr_data)

        await self._wait_for_cr_and_merge(cr_data)

        self.log.notice("Invoking completion handler for sub block with data {}".format(cr_data))
        completion_handler(cr_data)

        # Dev sanity check
        assert cr_data.input_hash in self.pending_futures, "Input hash {} removed from pending futures {}!"\
                                                           .format(cr_data.input_hash, self.pending_futures)

        if self.pending_futures[cr_data.input_hash]['merge']:
            self.log.debugv("Merge flag set to true for {}".format(cr_data))
            await self._wait_to_update_master_db(cr_data)
        else:
            self.pending_futures[cr_data.input_hash]['complete'] = True

    async def _wait_to_update_master_db(self, cr_data: CRContext):
        # If this is the first sub-block, then we are responsible for merging to master db. Thus we must wait
        # for everyone to finish conflict resolution before we merge to master.
        if self.sbb_idx == 0:
            await self._wait_for_everybody_cr(cr_data)
        self._update_master_db(expected_data=cr_data)

    async def _wait_for_available_db(self, input_hash: str):
        elapsed = 0
        while not (self._can_start_next_sb() and self.queued_futures[0]['input_hash'] == input_hash):
            await asyncio.sleep(Phase.POLL_INTERVAL)
            elapsed += Phase.POLL_INTERVAL

            if elapsed >= Phase.AVAIL_DB_TIMEOUT:
                err_msg = "Exceeded timeout of {} waiting for an available db for input hash {}!"\
                          .format(Phase.AVAIL_DB_TIMEOUT, input_hash)
                self.log.fatal(err_msg)
                raise Exception(err_msg)

        self.log.debugv("Waited a total of {} seconds for a db to become available for input hash {}"
                        .format(elapsed, input_hash))

    async def _wait_for_cr_and_merge(self, cr_data: CRContext):
        self.log.debug("Waiting for other SBBs to finish conflict resolution ({})...".format(cr_data))
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.CONFLICT_RESOLUTION, value=self.sbb_idx,
                                            timeout=(len(self.pending_dbs) + 1) * Phase.CR_TIMEOUT)
        self.log.debug("Done waiting for other SBBs to finish conflict resolution ({})".format(cr_data))

        self._rerun_contracts_for_cr_data(cr_data)
        self.log.info("Merging sbb_{} data to common layer ({})".format(self.sbb_idx, cr_data))
        cr_data.merge_to_common()

        self.log.debugv("Finished merging to common layer for CR {}. Incrementing CR Phase var.".format(cr_data))
        Phase.incr(cr_data.working_db, Macros.CONFLICT_RESOLUTION)

    async def _wait_until_top_of_pending(self, cr_data: CRContext):
        """ Waits until cr_data is at the top (first element) of self.pending_dbs """
        # TODO technically this 'wait' can be triggered reactively when we leftpop pending_dbs in merge_to_master
        # This is an optimization we can do later
        elapsed = 0
        timeout = Phase.BLOCK_TIMEOUT * len(self.pending_dbs)
        self.log.debug("CRData {} waiting its turn until its at the top of pending_dbs..."
                       .format(cr_data))

        while self.pending_dbs[0] is not cr_data:
            await asyncio.sleep(Phase.POLL_INTERVAL)
            elapsed += Phase.POLL_INTERVAL

            if elapsed >= timeout:
                err_msg = "Exceeded timeout of {} waiting for cr data {} to reach top of pending dbs!" \
                          .format(timeout, cr_data)
                self.log.fatal(err_msg)
                raise Exception(err_msg)

        self.log.debug("CRData {} DONE waiting to be at the top of pending db!".format(cr_data))

    async def _wait_for_execution_stage(self, cr_data: CRContext):
        self.log.debug("Waiting for other SBBs to finish execution ({})".format(cr_data))
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.EXECUTION, value=self.num_sb_builders,
                                            timeout=(len(self.pending_dbs) + 1) * Phase.EXEC_TIMEOUT)
        self.log.debug("Done waiting for other SBBs to finish execution ({})".format(cr_data))

    async def _wait_for_everybody_cr(self, cr_data: CRContext):
        self.log.debug("Waiting for conflict resolution to finish for ALL sub blocks ({})".format(cr_data))
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.CONFLICT_RESOLUTION,
                                            value=self.num_sb_builders, timeout=Phase.CR_TIMEOUT)
        self.log.debug("Conflict resolution complete for ALL sub blocks ({})".format(cr_data))

    async def _wait_for_phase_variable(self, db: redis.StrictRedis, key: str, value: int, timeout: int):
        elapsed = 0
        while Phase.get(db, key) != value:
            await asyncio.sleep(Phase.POLL_INTERVAL)
            elapsed += Phase.POLL_INTERVAL

            if elapsed >= timeout:
                err_msg = "Client with sbb_idx {} exceeded timeout of {} waiting for phase key {} to reach {}! " \
                          "Current conflict resolution phase value: {}" \
                    .format(self.sbb_idx, elapsed, key, value,
                            Phase.get(db, key))
                self.log.fatal(err_msg)
                raise Exception(err_msg)

        self.log.spam("Waited a total of {} seconds for phase variable {} to reach value {}".format(elapsed, key, value))

    def _ensure_future(self, coro) -> asyncio.Future:
        """
        A small wrapper around asyncio.ensure_future to catch any error and log them before raising them. We do this
        because sometimes asyncio does not properly raise the exceptions, causing coroutines to sometimes fail
        silently.
        """
        async def _safe_ensure_future():
            try:
                return await coro
            except Exception as e:
                if type(e) is asyncio.CancelledError:
                    self.log.warning("Coro {} cancelled.".format(coro))
                else:
                    self.log.fatal("\nError caught in coro {}!\n{}\n".format(coro, traceback.format_exc()))
                    raise e

        return asyncio.ensure_future(_safe_ensure_future())



from seneca.libs.datatypes import hmap

balances = hmap('balances', str, int)

@export
def transfer(to, amount):
    assert balances[rt['sender']] >= amount

    balances[to] += amount
    balances[rt['sender']] -= amount

# a better way to deal with 'allowances' which are dumb af
# and don't reflect real life business operations
custodials = hmap('custodials', str, hmap(key_type=str, value_type=int))

@export
def add_to_custodial(to, amount):
    assert balances[rt['sender']] >= amount

    custodials[rt['sender']][to] += amount
    balances[rt['sender']] -= amount

@export
def remove_from_custodial(to, amount):
    assert custodials[rt['sender']][to] >= amount

    balances[rt['sender']] += amount
    custodials[rt['sender']][to] -= amount

@export
def spend_custodial(_from, amount, to):
    assert custodials[_from][rt['sender']] >= amount

    balances[to] += amount
    custodials[_from][rt['sender']] -= amount

@export
def get_balance(account):
	return balances[account]

@export
def get_custodial(owner, spender):
	return custodials[owner][spender]

@seed
def seed():
	balances['stu'] = 1000000



import sys, redis
from contextlib import contextmanager
from io import StringIO
from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.constants.config import get_redis_port, MASTER_DB, DB_OFFSET, get_redis_password

def recur_fibo(n):
    if n <= 1:
        return n
    else:
        return(recur_fibo(n-1) + recur_fibo(n-2))

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestInterface(TestCase):

    def setUp(self):
        self.si = SenecaInterface(False, port=get_redis_port(), password=get_redis_password())
        try: v = self.si.r.get('market:stamps_to_tau')
        except: v = 1
        self.si.r.flushdb()
        self.si.r.set('market:stamps_to_tau', v)
        print('\n{}'.format('#' * 128))
        print(self.id)
        print('{}\n'.format('#' * 128))



from seneca.redis_importer import hmap, table
import hashlib

swaps = hmap('swaps', str, hmap(None, str, table(None, {
	'initiator': str,
	'participant': str,
	'amount': int,
	'token': str,
	'expiration': int,
})))

def initiate(_initiator: str,
	_participant: str,
	_expiration: int,
	_hashlock: bytes,
	_token: str,
	_amount: int):
	if not swaps[_participant][_hashlock].exists():

		# dynamic import of token
		token = importlib.import_module(_token)
		assert token.allowance(rt['sender'], rt['contract_address']) >= _amount
		token.transfer_from(rt['sender'], rt['contract_address'], _amount)

		swaps[_participant][_hashlock] = {
			'initiator': _initiator,
			'participant': _participant,
			'amount': _amount,
			'token': _token,
			'expiration': _expiration
		}

def redeem(secret: bytes):
	digest = hashlib.sha3_256().update(secret).digest()
	if swaps[rt['sender']][digest].exists():
		s = swaps[rt['sender']][digest]
		assert s['participant'] == rt['sender']

		token = importlib.import_module(s['token'])
		token.transfer(rt['sender'], s['value'])

def refund(secret: bytes):
	digest = hashlib.sha3_256().update(secret).digest()
	if swaps[rt['sender']][digest].exists():
		s = swaps[rt['sender']][digest]
		assert s['expiration'] < now
		assert s['initiator'] == rt['sender']

		token = importlib.import_module(s['token'])
		token.transfer(rt['sender'], s['value'])



from test_contracts.good import one_you_can_export as good
from test_contracts.okay import one_you_can_export as okay

@export
def good_call():
    good()
    okay()

@export
def reasonable_call():
    good()

@export
def do_that_thing():
    return 'sender: {}, author: {}'.format(rt['sender'], rt['author'])

@export
def test_global_namespace():
    print('sender: {}, author: {}'.format(rt['sender'], rt['author']))
    print("sbb_idx: {}".format(sbb_idx))
    print("ALL GLOBALS: {}".format(globals()))

def secret_call():
    okay()



from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from decimal import *

GENESIS_AUTHOR = 'davis'
STAMP_AMOUNT = None
MINT_WALLETS = {
    'davis': 10000,
    'stu': 69,
    'birb': 8000,
    'ghu': 9000,
    'tj': 8000,
    'ethan': 8000
}


class TestImporting(TestCase):
    CONTRACTS_TO_STORE = {
        'birb_bucks': 'birb_bucks.sen.py',
        'cat_cash': 'cat_cash.sen.py',
        'dynamic_imports': 'dynamic_imports.sen.py',
        'currency': 'currency.sen.py'}

    def setUp(self):
        # overwrite_logger_level(0)
        with SenecaInterface(False, 6379, '') as interface:
            interface.r.flushall()
            # Store all smart contracts in CONTRACTS_TO_STORE
            import seneca
            test_contracts_path = seneca.__path__[0] + '/../test_contracts/'

            for contract_name, file_name in self.CONTRACTS_TO_STORE.items():
                with open(test_contracts_path + file_name) as f:
                    code_str = f.read()
                    interface.publish_code_str(contract_name, GENESIS_AUTHOR, code_str)

            rt = {
                'author': GENESIS_AUTHOR,
                'sender': GENESIS_AUTHOR,
                'contract': 'minter'
            }

    def test_import(self):
        with SenecaInterface(False, 6379, '') as interface:
            f = interface.execute_function(
                module_path='seneca.contracts.dynamic_imports.get_token_balance',
                sender=GENESIS_AUTHOR,
                stamps=None,
                token_name='birb_bucks',
                account='birb'
            )

            self.assertEqual(f['output'], 1000000)

            f = interface.execute_function(
                module_path='seneca.contracts.dynamic_imports.get_token_balance',
                sender=GENESIS_AUTHOR,
                stamps=None,
                token_name='cat_cash',
                account='cat'
            )

            self.assertEqual(f['output'], 1000000)

            f = interface.execute_function(
                module_path='seneca.contracts.dynamic_imports.get_token_balance',
                sender=GENESIS_AUTHOR,
                stamps=None,
                token_name='cat_cash',
                account='birb'
            )

            self.assertEqual(f['output'], 0)



import os, threading
import redis
from multiprocessing import Lock
from seneca.engine.conflict_resolution import CRContext


class BookKeeper:
    _shared_state = {}
    _lock = Lock()

    @classmethod
    def _get_key(cls) -> str:
        """
        Returns a key unique to this particular thread/process combination.
        :return: The unique thead-process key (as a string)
        """
        key = "{}:{}".format(os.getpid(), threading.get_ident())
        return key

    @classmethod
    def _get_cr_key(cls) -> str:
        return cls._get_key() + ':CR'

    @classmethod
    def set_cr_info(cls, sbb_idx: int, contract_idx: int, data: CRContext, **kwargs) -> None:
        """
        Sets the info (subblock builder index and contract index) for the current thread.
        """
        key = cls._get_cr_key()
        with cls._lock:
            cls._shared_state[key] = {'sbb_idx': sbb_idx, 'contract_idx': contract_idx, 'data': data, **kwargs}

    @classmethod
    def get_cr_info(cls) -> dict:
        """
        Returns the info previously set for this specific thread by set_cr_info.
        :return:
        """
        key = cls._get_cr_key()
        with cls._lock:
            assert key in cls._shared_state, "Key {} not found in shared state. Did you call set_cr_info first?".format(key)
            return cls._shared_state[key]

    @classmethod
    def get_info(cls):
        key = cls._get_key()
        with cls._lock:
            assert key in cls._shared_state, "Key {} not found in shared state. Did you call set_info first?".format(key)
            return cls._shared_state[key]

    @classmethod
    def set_info(cls, **kwargs):
        key = cls._get_key()
        with cls._lock:
            cls._shared_state[key] = kwargs

    @classmethod
    def has_cr_info(cls) -> bool:
        """
        Checks if bookkeeping conflict resolution info exists for this current process/thread combination
        """
        key = cls._get_cr_key()
        with cls._lock:
            return key in cls._shared_state

    @classmethod
    def has_info(cls) -> bool:
        """
        Checks if bookkeeping info exists for this current process/thread combination
        """
        key = cls._get_key()
        with cls._lock:
            return key in cls._shared_state

    @classmethod
    def del_cr_info(cls) -> None:
        key = cls._get_cr_key()
        with cls._lock:
            assert key in cls._shared_state, "Key {} not found in shared state. Did you call set_cr_info first?"
            del cls._shared_state[key]

    @classmethod
    def del_info(cls) -> None:
        key = cls._get_key()
        with cls._lock:
            assert key in cls._shared_state, "Key {} not found in shared state. Did you call set_info first?"
            del cls._shared_state[key]

    @classmethod
    def reset(cls):
        with cls._lock:
            cls._shared_state.clear()



"""
    OperationsEstimator is for computing the estimated cost of each opcode
    instructions for your system. This is not meant to be used for determining
    the cost of a smart contract. For that, see seneca.libs.metering.cost.Cost
"""

import dis, time, os, sys, numpy as np
from collections import defaultdict
from scipy.optimize import nnls
import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt

import seneca

constants_path = seneca.__path__[0] + '/constants/'

def timeit(fn):
    def _fn(self, op, code_str, *args, **kwargs):

        ### Finding the median time it takes for each code snippets
        med = []
        opcode = dis.opmap[op]
        instructions = defaultdict(int)
        for ins in dis.get_instructions(code_str):
            instructions[ins.opcode] += 1
        code_obj = compile(code_str, '__main__', 'exec')
        for i in range(self.iterations):
            try:
                start = time.clock()
                fn(self, code_obj, *args, **kwargs)
                end = time.clock()
            except ValueError:
                end = time.clock()
            elapsed = end-start
            med.append(elapsed)
            if hasattr(self, 'histogram_file'):
                self.histogram_file.write('{},{}\n'.format(opcode, elapsed))

        t = np.median(med)
        # if not instructions.get(opcode):
        #     raise Exception('Not a valid example: {}\n{}'.format(op,
        #         [ins.opname for ins in dis.get_instructions(code_str)]))
        self.costs[opcode] = {
            't': t,
            'ops': instructions
        }

        return t
    return _fn

def suppress_stdout(fn):
    def _fn(*args, **kwargs):
        ### Disable stdout
        stdout = sys.stdout
        f = open(os.devnull, 'w')
        sys.stdout = f
        fn(*args, **kwargs)
        ### Re-enable stdout
        sys.stdout = stdout
    return _fn

class OperationsEstimator:
    def __init__(self, show_plot=False):
        self.iterations = 10000
        self.total_opcodes = 144
        self.non_zero_min_cost = np.float('2.7937841015807387e-09')
        self.costs = {}
        self.histogram_fname = os.path.join(constants_path, 'histogram.csv')
        self.costs_fname = os.path.join(constants_path, 'costs.csv')
        self.cu_costs_fname = os.path.join(constants_path, 'cu_costs.const')
        self.show_plot = show_plot
        if not os.path.exists(self.costs_fname):
            self.compute_opcode_costs()
        else:
            self.load_opcode_costs()
        if not os.path.exists(self.cu_costs_fname):
            self.comput_cu_costs()
        else:
            self.load_cu_costs()

    @timeit
    def calc_cost(self, code_obj):
        exec(code_obj)

    @suppress_stdout
    def precalc_costs(self):
        self.calc_cost('POP_TOP','''
(lambda:1)()
        ''')
        self.calc_cost('ROT_TWO','''
a = 1; b = 2
(a, b) = (b, a)
        ''')
        self.calc_cost('ROT_THREE','''
a = 1
(a, a, a) = (a, a, a)
        ''')
        self.calc_cost('UNARY_POSITIVE','''
a = 1
a = +a
        ''')
        self.calc_cost('UNARY_NEGATIVE','''
a = 1
a = -a
        ''')
        self.calc_cost('UNARY_NOT','''
a = 1
a = not a
        ''')
        self.calc_cost('UNARY_INVERT','''
a = 1
a = ~a
        ''')
        self.calc_cost('BINARY_POWER','''
a = 2
a = a ** 2
        ''')
        self.calc_cost('BINARY_MULTIPLY','''
a = 2
a = a * 2
        ''')
        self.calc_cost('BINARY_MODULO','''
a = 2
a = a % 2
        ''')
        self.calc_cost('BINARY_ADD','''
a = 2
a = a + 2
        ''')
        self.calc_cost('BINARY_SUBTRACT','''
a = 2
a = a - 2
        ''')
        self.calc_cost('BINARY_SUBSCR','''
a = [1]
a[0]
        ''')
        self.calc_cost('BINARY_FLOOR_DIVIDE','''
a = 2
a = a // 2
        ''')
        self.calc_cost('BINARY_TRUE_DIVIDE','''
a = 2
a = a / 2
        ''')
        self.calc_cost('INPLACE_FLOOR_DIVIDE','''
a = 1
a //= 1
        ''')
        self.calc_cost('BUILD_SLICE','''
a = [1,2,3]
a = a[:]
        ''')
        self.calc_cost('INPLACE_ADD','''
a = 1
a += 1
        ''')
        self.calc_cost('INPLACE_SUBTRACT','''
a = 1
a -= 1
        ''')
        self.calc_cost('INPLACE_MULTIPLY','''
a = 1
a *= 1
        ''')
        self.calc_cost('INPLACE_TRUE_DIVIDE','''
a = 1
a /= 1
        ''')
        self.calc_cost('INPLACE_MODULO','''
a = 1
a %= 1
        ''')
        self.calc_cost('STORE_SUBSCR','''
a = [0]
a[0] = 1
        ''')
        self.calc_cost('DELETE_SUBSCR','''
a = [1]
del a[0]
        ''')
        self.calc_cost('BINARY_LSHIFT','''
a = 1
a = a << 1
        ''')
        self.calc_cost('BINARY_RSHIFT','''
a = 1
a = a >> 1
        ''')
        self.calc_cost('BINARY_AND','''
a = 1
a = a & 1
        ''')
        self.calc_cost('BINARY_XOR','''
a = 1
a = a ^ 1
        ''')
        self.calc_cost('BINARY_OR','''
a = 1
a = a | 1
        ''')
        self.calc_cost('INPLACE_POWER','''
a = 1
a **= 1
        ''')
        self.calc_cost('GET_ITER','''
for a in (1,2): pass
        ''')
        self.calc_cost('INPLACE_LSHIFT','''
a = 1
a <<= 1
        ''')
        self.calc_cost('INPLACE_RSHIFT','''
a = 1
a >>= 1
        ''')
        self.calc_cost('INPLACE_AND','''
a = 1
a &= 1
        ''')
        self.calc_cost('INPLACE_XOR','''
a = 1
a ^= 1
        ''')
        self.calc_cost('INPLACE_OR','''
a = 1
a |= 1
        ''')
        self.calc_cost('BREAK_LOOP','''
for a in (1,2): break
        ''')
        f = open('1.txt', 'w+')
        self.calc_cost('WITH_CLEANUP_START','''
with open("1.txt") as f:
    print(f.read())
        ''')
        self.calc_cost('WITH_CLEANUP_FINISH','''
with open("1.txt") as f:
    print(f.read())
        ''')
        os.remove('1.txt')
        self.calc_cost('RETURN_VALUE','''
# empty file
        ''')
        self.calc_cost('IMPORT_STAR','''
from sys import *
        ''')
        self.calc_cost('POP_BLOCK','''
for a in (1,2): break
        ''')
        self.calc_cost('END_FINALLY','''
try:
    a = 1
finally:
    a = 3
        ''')
        self.calc_cost('LOAD_BUILD_CLASS','''
class a: pass
        ''')
        self.calc_cost('STORE_NAME','''
a = 1
        ''')
        self.calc_cost('DELETE_NAME','''
a = 1
del a
        ''')
        self.calc_cost('UNPACK_SEQUENCE','''
(a, b) = "ab"
        ''')
        self.calc_cost('FOR_ITER','''
for i in (1,2): pass
        ''')
        self.calc_cost('STORE_ATTR','''
import sys
sys.stderr = sys.stdout
        ''')
        self.calc_cost('DELETE_ATTR','''
import random
random.hello = 'world'
del random.hello
        ''')
        self.calc_cost('STORE_GLOBAL','''
global a
a = 1
        ''')
        self.calc_cost('DELETE_GLOBAL','''
global a
a = 2
del a
        ''')
        self.calc_cost('DUP_TOP_TWO','''
a = 0
b = [0]
b[a] += 1
        ''')
        self.calc_cost('LOAD_CONST','''
a = 1
        ''')
        self.calc_cost('LOAD_NAME','''
a = 1
a = a
        ''')
        self.calc_cost('BUILD_TUPLE','''
a = 1;
a = (a, a)
        ''')
        self.calc_cost('BUILD_LIST','''
[1,2,3]
        ''')
        self.calc_cost('BUILD_CONST_KEY_MAP','''
{"a":1,"b":2}
        ''')
        self.calc_cost('LOAD_ATTR','''
[].sort()
        ''')
        self.calc_cost('COMPARE_OP','''
a = 1 == 2
        ''')
        self.calc_cost('IMPORT_NAME','''
import random
        ''')
        self.calc_cost('IMPORT_FROM','''
from dis import opmap
        ''')
        self.calc_cost('JUMP_FORWARD','''
if 1 == 2: pass
else: pass
        ''')
        self.calc_cost('POP_JUMP_IF_FALSE','''
if 1 == 2: pass
else: pass
        ''')
        self.calc_cost('POP_JUMP_IF_TRUE','''
if not(1 == 2): pass
else: pass
        ''')
        self.calc_cost('JUMP_ABSOLUTE','''
for i in (1,2): pass
        ''')
        self.calc_cost('LOAD_GLOBAL','''
global a
a = 1
a = a
        ''')
        self.calc_cost('CONTINUE_LOOP','''
for x in (1,2):
    try: continue
    except: pass
        ''')
        self.calc_cost('SETUP_LOOP','''
while 0 > 1: pass
        ''')
        self.calc_cost('SETUP_EXCEPT','''
try:
    a = 1
except ValueError:
    a = 2
        ''')
        self.calc_cost('SETUP_FINALLY','''
try:
    a = 1
finally:
    a = 3
        ''')
        self.calc_cost('RAISE_VARARGS','''
raise ValueError
        ''')
        self.calc_cost('CALL_FUNCTION','''
def f(): pass
f()
        ''')
        self.calc_cost('MAKE_FUNCTION','''
def f(): pass
        ''')
        self.calc_cost('BUILD_SLICE','''
a = [1,2,3,4]
b = a[::-1]
        ''')
        self.calc_cost('CALL_FUNCTION_EX','''
def f(a,b): pass
a = (1,2)
f(*a)
        ''')

    def computer_medians(self):
        X = []
        y = []
        for opcode, item in self.costs.items():
            X.append([item['ops'][i] if item['ops'].get(i) else 0 for i in range(self.total_opcodes)])
            y.append(item['t'])
        X = np.array(X)
        y = np.array(y)
        medians = np.array(nnls(X, y)[0])
        self.opcodes = [i for i in range(self.total_opcodes)]
        self.medians = [m for idx, m in enumerate(medians)]
        self.costs = {}
        for i, m in enumerate(self.medians):
            self.costs_file.write('{},{}\n'.format(self.opcodes[i], repr(m)))

    def plot(self):
        plt.plot(self.opcodes, self.medians, 'o', markersize=5)
        plt.legend()
        plt.show()

    def compute_opcode_costs(self):
        self.histogram_file = open(self.histogram_fname, 'w+')
        self.costs_file = open(self.costs_fname, 'w+')
        self.precalc_costs()
        self.computer_medians()
        if self.show_plot: self.plot()

    def load_opcode_costs(self):
        self.medians = []
        with open(self.costs_fname) as f:
            for line in f:
                opcode, median = line.split(',')
                self.medians.append(np.float(median))

    def comput_cu_costs(self):
        self.cu_costs_file = open(self.cu_costs_fname, 'w+')
        self.cu_costs = {}
        costs = np.array(self.medians)/self.non_zero_min_cost + 1.0 # 1.0 is padding for 0 costs
        for opcode, m in enumerate(costs):
            self.cu_costs[opcode] = int(np.round(m))
            self.cu_costs_file.write('{},{}\n'.format(opcode, self.cu_costs[opcode]))

    def load_cu_costs(self):
        self.cu_costs = {}
        with open(self.cu_costs_fname) as f:
            for line in f:
                opcode, cost = line.split(',')
                self.cu_costs[int(opcode)] = int(cost)



import nacl
import nacl.encoding
import nacl.signing


def verify(v: bytes, msg: bytes, sig: bytes):
    v = nacl.signing.VerifyKey(v)
    try:
        v.verify(msg, sig)
    except nacl.exceptions.BadSignatureError:
        return False
    return True



import random
from seneca.engine.book_keeper import BookKeeper

'''
    This module wraps and exposes the Python stdlib random functions that can be made deterministic with a random seed 
    and return fixed point precision where possible. This allows some psuedorandom behavior when it is nice to have, but
    with the caveat that it's based on environmental constants such as the last block hash and public information such
    as the sender's address to seed the random state so it's not *really* random.
    
    It's most likely 'random enough' for most cases, but people can always theoretically reproduce the seed and try to
    front-run a smart contract by testing the seeded randoms for a preferable outcome and submitting a transaction
    before the next block is minted. While this is extremely unlikely and hard to pull off, it's a valid hole in the
    security and needs to be accepted as a flaw when using random numbers on a reproducible transaction log such as a 
    blockchain.
'''


class Seeded:
    s = False


def seed(d=None):
    try:
        seed = BookKeeper.get_cr_info().get('last_block_hash')
    except Exception as e:
        seed = b'0'

    if isinstance(seed, str):
        seed.encode()

    assert isinstance(d, bytes) or d is None, 'Must provide bytes to seed.'

    if d is not None:
        seed += d

    random.seed(seed)
    Seeded.s = True


def getrandbits(k):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    return random.getrandbits(k)


def shuffle(l):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    random.shuffle(l)


def randrange(k):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    return random.randrange(k)


def randint(a, b):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    return random.randint(a, b)


def choice(l):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    return random.choice(l)


def choices(l, k):
    assert Seeded.s, 'Random state not seeded. Call seed().'
    return random.choices(l, k=k)



from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter, ReadOnlyException, CompilationException
from os.path import join
from tests.utils import captured_output, TestInterface
import redis, unittest, seneca

test_contracts_path = seneca.__path__[0] + '/test_contracts/'

class TestImports(TestInterface):

    def test_import_valid(self):
        """
            This is a valid way to import
        """
        self.si.execute_code_str("""
from test_contracts.sample import good_call, reasonable_call
        """)

    def test_import_invalid(self):
        """
            This is a valid way to import, but you cannot import "importlib"
            and other such libraries. Only ones from the whitelist
        """
        with self.assertRaises(ImportError) as context:
            self.si.execute_code_str("""
from test_contracts.bad import innocent_function
            """)

    def test_import_direct_invalid(self):
        """
            This is a valid way to import, but you cannot import "importlib"
            and other such libraries. Only ones from the whitelist
        """
        with self.assertRaises(ImportError) as context:
            self.si.execute_code_str("""
import json
            """)

    def test_import_global_variable_invalid(self):
        """
            This is a valid way to import, but you cannot import "importlib"
            and other such libraries. Only ones from the whitelist
        """

        with captured_output() as (out, err):
            with self.assertRaises(CompilationException) as context:
                self.si.execute_code_str("""
from test_contracts.good import balances
print('Hacked', balances) # Should not print this!
                """)
            self.assertEqual(out.getvalue().strip(), '')

    def test_import_star(self):
        """
            Import * is currently not allowed
        """
        with self.assertRaises(ImportError) as context:
            self.si.execute_code_str("""
from test_contracts.sample import *
            """)

    def test_import_indirect_invalid_import(self):
        """
            You cannot import the entire module. You must import its functions
            seperately. (See TestInterface.test_import_valid())
        """
        with self.assertRaises(ImportError) as context:
            self.si.execute_code_str("""
import test_contracts.sample
            """)

    def test_import_indirect_invalid_from_import(self):
        """
            Just as the previous test case, you cannot import the whole module
            using this syntax neither.
        """
        with self.assertRaises(ImportError) as context:
            self.si.execute_code_str("""
from test_contracts import sample
            """)

    def test_execute_valid(self):
        """
            Testing to see if the function can be called.
        """
        self.si.execute_code_str("""
from test_contracts.sample import good_call
good_call()
        """)

if __name__ == '__main__':
    unittest.main()



from seneca.engine.interface import SenecaInterface
import types


class LocalInterface(SenecaInterface):
    def __init__(self, concurrent_mode=False,
                       port=6379,
                       password='',
                       bypass_currency=True):

        super().__init__(concurrent_mode=concurrent_mode,
                         port=port,
                         password=password,
                         bypass_currency=bypass_currency)

    def delete_contract(self, name):
        self.r.hdel('contracts', name)
        self.r.hdel('contracts_code', name)
        self.r.hdel('contracts_meta', name)

        for key in self.r.scan_iter('{}:*'.format(name)):
            self.r.delete(key)


default_driver = LocalInterface(concurrent_mode=False,
                                port=6379,
                                password='',
                                bypass_currency=True)


class SenecaFunction:
    def __init__(self, name, module_path, kwargs, default_sender, driver):
        self.name = name
        self.module_path = module_path
        self.kwargs = kwargs
        self.defaults = {
            'sender': default_sender
        }
        self.driver = driver

    def __call__(self, *args, **kwargs):

        def default(d, k):
            return d if kwargs.get(k) is None else kwargs.get(k)

        sender = self.defaults.get('sender')
        if 'sender' in kwargs.keys():
            sender = kwargs['sender']
            kwargs.pop('sender', None)

        stamps = default(None, 'stamps')

        kwargs['stamps'] = stamps
        kwargs['sender'] = sender

        r = self.driver.execute_function(
            module_path=self.module_path,
            **kwargs
        )

        return r


class ContractWrapper:
    def __init__(self, contract_name=None, driver=default_driver, default_sender=None):
        self.driver = driver
        self.author = driver.get_contract_meta(contract_name)['author']
        self.default_sender = default_sender
        contract_code = driver.get_code_obj(contract_name)

        codes = [cd for cd in contract_code.co_consts if type(cd) == types.CodeType]
        for _c in codes:
            name = _c.co_name
            module_path = 'seneca.contracts.{}.{}'.format(contract_name, _c.co_name)
            kwargs = _c.co_varnames
            setattr(self, name, SenecaFunction(name=name,
                                               module_path=module_path,
                                               kwargs=kwargs,
                                               default_sender=self.default_sender,
                                               driver=self.driver))


def publish_function(f, name, author):
    default_driver.publish_function(f, contract_name=name, author=author)
    return ContractWrapper(contract_name=name, driver=default_driver, default_sender=author)


def export(*args):
    pass



from unittest import TestCase
from unittest import mock
from unittest.mock import MagicMock
from seneca.engine.client import *
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter
from seneca.libs.logger import overwrite_logger_level, get_logger
from decimal import Decimal
from collections import OrderedDict, defaultdict
import random, uuid


log = get_logger("TestSenecaClient")
GENESIS_AUTHOR = 'anonymoose'
STAMP_AMOUNT = None
MINT_WALLETS = {
    'anonymoose': 10000,
    'stu': 69,
    'birb': 8000,
    'ghu': 9000,
    'tj': 8000,
    'ethan': 8000
}

# Add a bunch of other random wallet
for _ in range(359):
    MINT_WALLETS[str(uuid.uuid4())] = 2 ** 63


TEST_CONTRACT = \
"""
balances = {'hello': 'world'}
@export
def one_you_can_export():
    print('Running one_you_can_export()')
@export
def one_you_can_also_export():
    print('Running one_you_can_also_export()')
    one_you_can_export()
def one_you_cannot_export(dont, do, it='wrong'):
    print('Always runs: Running one_you_cannot_export()')
@export
def one_you_can_also_also_export():
    print('Running one_you_can_also_also_export()')
    one_you_cannot_export('a', 'b', it='c')
"""


class MockContractTransaction:
    def __init__(self, sender: str, contract_name: str, func_name: str, stamps=STAMP_AMOUNT, **kwargs):
        self.stamps_supplied, self.sender, self.func_name, self.contract_name = stamps, sender, func_name, contract_name
        self.kwargs = kwargs


class MockPublishTransaction:
    def __init__(self, sender: str, contract_name: str, contract_code: str, stamps=STAMP_AMOUNT):
        self.stamps_supplied, self.sender, self.contract_code, self.contract_name = stamps, sender, contract_code, contract_name


def create_currency_tx(sender: str, receiver: str, amount: int, contract_name: str='currency', stamps=STAMP_AMOUNT):
    contract = MockContractTransaction(sender=sender, contract_name='currency', func_name='transfer', stamps=stamps,
                                       to=receiver, amount=amount)
    return contract


class TestSenecaClient(TestCase):
    LOG_LVL = None
    CONTRACTS_TO_STORE = {'currency': 'currency.sen.py'}

    @classmethod
    def setUpClass(cls):
        if cls.LOG_LVL:
            overwrite_logger_level(cls.LOG_LVL)

    @classmethod
    def tearDownClass(cls):
        if cls.LOG_LVL:
            overwrite_logger_level(999999)  # re-enable all logging

    def assert_completion(self, expected_sbb_rep: List[tuple]=None, input_hash='', merge_master=False, client=None, merge_wait=1):
        if merge_master:
            assert client is not None, "if merge_master=True then client must be passed in"

        async def _merge(client):
            asyncio.sleep(merge_wait)
            client.update_master_db()

        def _completion_handler(cr_data: CRContext):
            if input_hash:
                self.assertEqual(cr_data.input_hash, input_hash)
            if expected_sbb_rep:
                self.assertEqual(expected_sbb_rep, cr_data.get_subblock_rep())
            if merge_master:
                asyncio.ensure_future(_merge(client))
            if client and input_hash:
                self.completed_hashes[client].append(input_hash)

        return _completion_handler

    def _get_futures(self, input_hash_client_dict: dict) -> list:
        futs = []
        for input_hash, client in input_hash_client_dict.items():
            if input_hash in client.pending_futures:
                d = client.pending_futures[input_hash]
                futs.append(d['fut'])
                if d['merge_fut']:
                    futs.append(d['merge_fut'])
            for data in client.queued_futures:
                if data['input_hash'] == input_hash:
                    futs.append(data['fut'])

        return futs

    def _gen_random_contracts(self, num=8, max_amount=8, stamps=STAMP_AMOUNT) -> list:
        contracts = []
        for _ in range(num):
            sender, receiver = random.sample(list(MINT_WALLETS.keys()), 2)
            amount = random.randint(0, max_amount)
            contracts.append(create_currency_tx(sender, receiver, amount, stamps=stamps))

        return contracts

    def _mint_wallets(self, seed_amount=None):
        with SenecaInterface(False) as interface:
            for wallet, amount in MINT_WALLETS.items():
                interface.execute_function(module_path='seneca.contracts.currency.mint', sender=GENESIS_AUTHOR,
                                           stamps=STAMP_AMOUNT, to=wallet, amount=seed_amount or amount)

    def setUp(self):
        # overwrite_logger_level(0)
        with SenecaInterface(False) as interface:
            interface.r.flushall()
            # Store all smart contracts in CONTRACTS_TO_STORE
            import seneca
            test_contracts_path = seneca.__path__[0] + '/../test_contracts/'

            for contract_name, file_name in self.CONTRACTS_TO_STORE.items():
                with open(test_contracts_path + file_name) as f:
                    code_str = f.read()
                    interface.publish_code_str(contract_name, GENESIS_AUTHOR, code_str)

            rt = {
                'sender': GENESIS_AUTHOR,
                'contract': 'minter'
            }

        self._mint_wallets()
        self.completed_hashes = defaultdict(list)

    def test_setup_dbs(self):
        client = SenecaClient(sbb_idx=0, num_sbb=1)

        self.assertTrue(client.master_db is not None)
        self.assertTrue(client.active_db is None)

        self.assertEqual(len(client.available_dbs), NUM_CACHES)

    def test_flush(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)

        client._start_sb('A' * 64)
        client.run_contract(c1)
        client.run_contract(c2)

        client.flush_all()

        self.assertEqual(len(client.pending_dbs), 0)
        self.assertEqual(client.active_db, None)

        loop.close()

    def test_skip_current_db(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        input_hash = 'A' * 64
        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)

        client.execute_sb(input_hash=input_hash, contracts=[c1, c2],
                          completion_handler=self.assert_completion(None, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        client.skip_current_db()
        self.assertEqual(len(client.pending_futures), 0)

        loop.close()

    def test_run_tx_increments_contract_idx(self):
        client = SenecaClient(sbb_idx=0, num_sbb=1)
        client._start_sb('A' * 64)

        self.assertEqual(client.active_db.next_contract_idx, 0)

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)

        client.run_contract(c1)
        self.assertEqual(client.active_db.next_contract_idx, 1)

        client.run_contract(c2)
        self.assertEqual(client.active_db.next_contract_idx, 2)

    def test_with_publish_transactions(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = MockPublishTransaction(sender='anonymoose', contract_name='test', contract_code=TEST_CONTRACT)
        expected_sbb_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                            (c2, "SUCC", "")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client._start_sb(input_hash)

        client.run_contract(c1)
        client.run_contract(c2)

        client._end_sb(self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_end_subblock_1_sbb(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                            (c2, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client._start_sb(input_hash)

        client.run_contract(c1)
        client.run_contract(c2)

        client._end_sb(self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_execute_sb(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                            (c2, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)

        client.execute_sb(input_hash=input_hash, contracts=[c1, c2],
                          completion_handler=self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_execute_empty_sb(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        expected_sbb_rep = []
        input_hash = 'A' * 64

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client.execute_sb(input_hash=input_hash, contracts=[],
                          completion_handler=self.assert_completion(expected_sbb_rep, input_hash))

        loop.run_until_complete(client.pending_futures[input_hash]['fut'])
        loop.close()

    def test_two_sb_one_empty_one_not(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        expected_sbb1_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                             (c2, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]
        expected_sbb2_rep = []

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)
        client1.execute_sb(input_hash=input_hash1, contracts=[c1, c2], completion_handler=self.assert_completion(expected_sbb1_rep, input_hash1))
        client2.execute_sb(input_hash=input_hash2, contracts=[], completion_handler=self.assert_completion(expected_sbb2_rep, input_hash2))

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))
        loop.close()

    def test_end_subblock_1_sbb_with_failure(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash = 'A' * 64
        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 9000)
        c3 = create_currency_tx('stu', 'anonymoose', 40)
        expected_sbb_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                            (c2, "FAIL -- Sender balance must be non-negative!!!", ""),
                            (c3, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]

        client = SenecaClient(sbb_idx=0, num_sbb=1, loop=loop)
        client._start_sb(input_hash)

        client.run_contract(c1)
        client.run_contract(c2)
        client.run_contract(c3)

        client._end_sb(self.assert_completion(expected_sbb_rep, input_hash))
        self.assertTrue(input_hash in client.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        loop.run_until_complete(client.pending_futures[input_hash]['fut'])

        loop.close()

    def test_end_subblock_2_sbb(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        c3 = create_currency_tx('ghu', 'anonymoose', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        expected_sbb1_rep = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                             (c2, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]
        expected_sbb2_rep = [(c3, "SUCC", "SET currency:balances:ghu 8985;SET currency:balances:anonymoose 10041;"),
                             (c4, "SUCC", "SET currency:balances:tj 7910;SET currency:balances:birb 8090;")]

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)
        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(expected_sbb1_rep, input_hash1))
        client2._end_sb(self.assert_completion(expected_sbb2_rep, input_hash2))
        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        loop.close()

    def test_end_subblock_2_sbb_start_subblocks_before_ending_then_flush_all_dem_bois(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64
        input_hash3 = 'C' * 64
        input_hash4 = 'D' * 64

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        c3 = create_currency_tx('ghu', 'anonymoose', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        c5 = create_currency_tx('ethan', 'birb', 60)
        c6 = create_currency_tx('stu', 'anonymoose', 10)
        c7 = create_currency_tx('ghu', 'tj', 50)
        c8 = create_currency_tx('birb', 'anonymoose', 100)
        expected_sbb1_1 = [(c1, "SUCC", "SET currency:balances:anonymoose 9986;SET currency:balances:stu 83;"),
                           (c2, "SUCC", "SET currency:balances:stu 43;SET currency:balances:anonymoose 10026;")]
        expected_sbb2_1 = [(c3, "SUCC", "SET currency:balances:ghu 8985;SET currency:balances:anonymoose 10041;"),
                           (c4, "SUCC", "SET currency:balances:tj 7910;SET currency:balances:birb 8090;")]
        expected_sbb1_2 = [(c5, "SUCC", "SET currency:balances:ethan 7940;SET currency:balances:birb 8150;"),
                           (c6, "SUCC", "SET currency:balances:stu 33;SET currency:balances:anonymoose 10051;")]
        expected_sbb2_2 = [(c7, "SUCC", "SET currency:balances:ghu 8935;SET currency:balances:tj 7960;"),
                           (c8, "SUCC", "SET currency:balances:birb 8050;SET currency:balances:anonymoose 10151;")]

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(expected_sbb1_1, input_hash1))
        client2._end_sb(self.assert_completion(expected_sbb2_1, input_hash2))

        client1._start_sb(input_hash3)
        client2._start_sb(input_hash4)

        client1.run_contract(c5)
        client1.run_contract(c6)
        client2.run_contract(c7)
        client2.run_contract(c8)

        client1._end_sb(self.assert_completion(expected_sbb1_2, input_hash3))
        client2._end_sb(self.assert_completion(expected_sbb2_2, input_hash4))

        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash3 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)
        self.assertTrue(input_hash4 in client2.pending_futures)

        client1.update_master_db()
        client2.update_master_db()

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        coros = (client1.pending_futures[input_hash3]['fut'], client2.pending_futures[input_hash4]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        loop.close()

    def test_update_master_db_with_incomplete_sb(self):
        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64
        input_hash3 = 'C' * 64
        input_hash4 = 'D' * 64

        c1 = create_currency_tx('anonymoose', 'stu', 14)
        c2 = create_currency_tx('stu', 'anonymoose', 40)
        c3 = create_currency_tx('ghu', 'anonymoose', 15)
        c4 = create_currency_tx('tj', 'birb', 90)
        c5 = create_currency_tx('ethan', 'birb', 60)
        c6 = create_currency_tx('stu', 'anonymoose', 10)
        c7 = create_currency_tx('ghu', 'tj', 50)
        c8 = create_currency_tx('birb', 'anonymoose', 100)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        client1._start_sb(input_hash1)
        client2._start_sb(input_hash2)

        client1.run_contract(c1)
        client1.run_contract(c2)
        client2.run_contract(c3)
        client2.run_contract(c4)

        client1._end_sb(self.assert_completion(None, input_hash1))
        client2._end_sb(self.assert_completion(None, input_hash2))

        client1._start_sb(input_hash3)
        client2._start_sb(input_hash4)

        client1.run_contract(c5)
        client1.run_contract(c6)
        client2.run_contract(c7)
        client2.run_contract(c8)

        client1._end_sb(self.assert_completion(None, input_hash3))
        client2._end_sb(self.assert_completion(None, input_hash4))

        self.assertTrue(input_hash1 in client1.pending_futures)
        self.assertTrue(input_hash3 in client1.pending_futures)
        self.assertTrue(input_hash2 in client2.pending_futures)
        self.assertTrue(input_hash4 in client2.pending_futures)

        client1.update_master_db()
        client2.update_master_db()

        # We must run the future manually, since the event loop is not currently running
        # coros1 = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        coros1 = self._get_futures({input_hash1: client1, input_hash2: client2})
        # coros1 = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'],
        #          client1.pending_futures[input_hash3]['fut'])
        loop.run_until_complete(asyncio.gather(*coros1))

        client2.update_master_db()
        client1.update_master_db()

        # coros2 = (client1.pending_futures[input_hash3]['fut'], client2.pending_futures[input_hash4]['fut'])
        coros2 = self._get_futures({input_hash3: client1, input_hash4: client2})
        loop.run_until_complete(asyncio.gather(*coros2))

        for c in (client1, client2):
            all_input_hashes = [cr.input_hash for cr in c.pending_dbs]
            self.assertEqual(len(c.pending_dbs), 0, "Client with sbb_idx {} has pending_dbs: {}".format(c.sbb_idx, all_input_hashes))

        loop.close()

    @mock.patch("seneca.engine.client.NUM_CACHES", 2)
    def test_hella_subblocks_with_stamps(self):
        self._mint_wallets(10 ** 8)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash1 = '1' * 64
        input_hash2 = '2' * 64
        input_hash3 = '3' * 64
        input_hash4 = '4' * 64
        input_hash5 = '5' * 64
        input_hash6 = '6' * 64
        input_hash7 = '7' * 64
        input_hash8 = '8' * 64
        input_hash9 = 'A' * 64
        input_hash10 = 'B' * 64

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        c1_map = OrderedDict({input_hash1: client1, input_hash3: client1, input_hash5: client1, input_hash7: client1})
        c2_map = OrderedDict({input_hash2: client2, input_hash4: client2, input_hash6: client2, input_hash8: client2})

        NUM_TX = 10
        for i, in_hash in enumerate(c1_map):
            # txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if i % 2 == 1 else []
            txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if True else []
            client1.execute_sb(in_hash, txs, self.assert_completion(None, in_hash, merge_master=True, client=client1, merge_wait=0))
        for i, in_hash in enumerate(c2_map):
            # txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if i % 2 == 1 else []
            txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if True else []
            client2.execute_sb(in_hash, txs, self.assert_completion(None, in_hash, merge_master=True, client=client2, merge_wait=0))

        # Execute an empty sb at the end
        client1.execute_sb(input_hash9, [], self.assert_completion(None, input_hash9, merge_master=True, client=client1, merge_wait=0))
        client2.execute_sb(input_hash10, [], self.assert_completion(None, input_hash10, merge_master=True, client=client2, merge_wait=0))

        # Run it all
        coros1 = self._get_futures(c1_map)
        coros2 = self._get_futures(c2_map)

        loop.run_until_complete(asyncio.gather(*coros1, *coros2))

        async def _wait_for_things_to_finish():
            log.notice("Tester waiting for things to finish")
            await asyncio.sleep(3)
            log.notice("Tester done waiting")

        loop.run_until_complete(_wait_for_things_to_finish())
        loop.close()

    @mock.patch("seneca.engine.client.NUM_CACHES", 2)
    def test_hella_subblocks_called_in_correct_order(self):
        self._mint_wallets(10 ** 8)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        input_hash1 = '1' * 64
        input_hash2 = '2' * 64
        input_hash3 = '3' * 64
        input_hash4 = '4' * 64
        input_hash5 = '5' * 64
        input_hash6 = '6' * 64
        input_hash7 = '7' * 64
        input_hash8 = '8' * 64
        input_hash9 = 'A' * 64
        input_hash10 = 'B' * 64

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        c1_map = OrderedDict({input_hash1: client1, input_hash3: client1, input_hash5: client1, input_hash7: client1})
        c2_map = OrderedDict({input_hash2: client2, input_hash4: client2, input_hash6: client2, input_hash8: client2})

        NUM_TX = 23
        for i, in_hash in enumerate(c1_map):
            # txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if i % 2 == 1 else []
            txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if True else []
            client1.execute_sb(in_hash, txs, self.assert_completion(None, in_hash, merge_master=True, client=client1, merge_wait=0))
        for i, in_hash in enumerate(c2_map):
            # txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if i % 2 == 1 else []
            txs = self._gen_random_contracts(num=NUM_TX, stamps=10 ** 5) if True else []
            client2.execute_sb(in_hash, txs, self.assert_completion(None, in_hash, merge_master=True, client=client2, merge_wait=0))

        # Execute an empty sb at the end
        client1.execute_sb(input_hash9, [], self.assert_completion(None, input_hash9, merge_master=True, client=client1, merge_wait=0))
        client2.execute_sb(input_hash10, [], self.assert_completion(None, input_hash10, merge_master=True, client=client2, merge_wait=0))

        # Run it all
        coros1 = self._get_futures(c1_map)
        coros2 = self._get_futures(c2_map)

        loop.run_until_complete(asyncio.gather(*coros1, *coros2))

        async def _wait_for_things_to_finish():
            log.notice("Tester waiting for things to finish")
            await asyncio.sleep(3)
            log.notice("Tester done waiting")

        loop.run_until_complete(_wait_for_things_to_finish())
        loop.close()

        # Check things were called in the correct order
        self.assertEqual(list(c1_map.keys()) + [input_hash9], self.completed_hashes[client1])
        self.assertEqual(list(c2_map.keys()) + [input_hash10], self.completed_hashes[client2])

    def test_contract_with_key_that_doesnt_exist_yet(self):
        self._mint_wallets(10 ** 8)

        input_hash1 = 'A' * 64
        input_hash2 = 'B' * 64
        input_hash3 = 'C' * 64
        input_hash4 = 'D' * 64

        stamps = 10**4
        c1 = create_currency_tx('anonymoose', 'stu', 14, stamps=stamps)
        c2 = create_currency_tx('stu', 'anonymoose', 40, stamps=stamps)
        c3 = create_currency_tx('ghu', 'A NEW KEY THAT DOESNT EXIST', 10**9, stamps=stamps)
        c4 = create_currency_tx('tj', 'birb', 90, stamps=stamps)
        c5 = create_currency_tx('ethan', 'birb', 60, stamps=stamps)
        c6 = create_currency_tx('stu', 'anonymoose', 10, stamps=stamps)
        c7 = create_currency_tx('ghu', 'tj', 50, stamps=stamps)
        c8 = create_currency_tx('birb', 'anonymoose', 100, stamps=stamps)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        client1 = SenecaClient(sbb_idx=0, num_sbb=2, loop=loop)
        client2 = SenecaClient(sbb_idx=1, num_sbb=2, loop=loop)

        client1.execute_sb(input_hash1, contracts=[c1, c2], completion_handler=self.assert_completion(None, input_hash1))
        client2.execute_sb(input_hash2, contracts=[c3, c4], completion_handler=self.assert_completion(None, input_hash2))

        client1.execute_sb(input_hash3, contracts=[c5, c6], completion_handler=self.assert_completion(None, input_hash3))
        client2.execute_sb(input_hash4, contracts=[c7, c8], completion_handler=self.assert_completion(None, input_hash4))

        client1.update_master_db()
        client2.update_master_db()

        # We must run the future manually, since the event loop is not currently running
        coros = (client1.pending_futures[input_hash1]['fut'], client2.pending_futures[input_hash2]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        coros = (client1.pending_futures[input_hash3]['fut'], client2.pending_futures[input_hash4]['fut'])
        loop.run_until_complete(asyncio.gather(*coros))

        loop.close()

    # Test with multiple sb's where stuff in SB 2 will pass the first time and fail the second time (cause some og read was modified)


if __name__ == "__main__":
    import unittest
    unittest.main()



from seneca.libs.datatypes import hmap

balances = hmap('balances', str, int)
allowed = hmap('allowed', str, hmap(value_type=int))
market = hmap('market', str, int)

@seed
def deposit_to_all_wallets():
    market['stamps_to_tau'] = 1
    balances['black_hole'] = 0

    STU = ('db929395f15937f023b4682995634b9dc19b1a2b32799f1f67d6f080b742cdb1',
           '324ee2e3544a8853a3c5a0ef0946b929aa488cbe7e7ee31a0fef9585ce398502')
    DAVIS = ('21fee38471799f8c2989dd81c6d46f6c2e2db6caf63efa98a093fcba064a4b62',
             'a103715914a7aae8dd8fddba945ab63a169dfe6e37f79b4a58bcf85bfd681694')
    DENTON = ('9decc7f7f0b5a4fc87ab5ce700e2d6c5d51b7565923d50ea13cbf78031bb3acf',
              '20da05fdba92449732b3871cc542a058075446fedb41430ee882e99f9091cc4d')
    FALCON = ('bac886e7c6e4a9fae572e170adb333b27b590157409e62d88cc0c7bc9a7b3631',
              'ed19061921c593a9d16875ca660b57aa5e45c811c8cf7af0cfcbd23faa52cbcd')
    CARL = ('cf67a180f9578afa5fd704cea39b450c1542755d73614f6a4f41b627190b83bb',
            'cb9bfd4b57b243248796e9eb90bc4f0053d78f06ce68573e0fdca422f54bb0d2')
    RAGHU = ('b44a8cc3dcadbdb3352ea046ec85cd0f6e8e3f584e3d6eb3bd10e142d84a9668',
             'c1f845ad8967b93092d59e4ef56aef3eba49c33079119b9c856a5354e9ccdf84')

    SEED_AMOUNT = 2 ** 63
    ALL_WALLETS = [STU, DAVIS, DENTON, FALCON, CARL, RAGHU]

    # this is hella silly lmao look what i have to do to just to seed a list of wallets. we should really just allow
    # for loops, i see little point in disallowing them
    def seed_wallet(wallet):
        print('seeding wallet {}'.format(wallet[1]))
        balances[wallet[1]] = SEED_AMOUNT
    list(map(seed_wallet, ALL_WALLETS))

    balances[STU[1]] = SEED_AMOUNT
    assert balances.get(STU[1]) == SEED_AMOUNT, "Seeding did not work!"
    assert balances.get(DAVIS[1]) == SEED_AMOUNT, "Seeding did not work!"

@export
def submit_stamps(stamps):
    # Special seeding case because it's stamps!
    if market['stamps_to_tau'] == 0:
        balances['anonymoose'] = 2 ** 63
        market['stamps_to_tau'] = 1
    amount = stamps / market['stamps_to_tau']
    transfer('black_hole', int(amount))

@export
def balance_of(wallet_id):
    return balances[wallet_id]

@export
def transfer(to, amount):
    # print("transfering from {} to {} with amount {}".format(rt['sender'], to, amount))
    balances[rt['sender']] -= amount
    balances[to] += amount
    sender_balance = balances[rt['sender']]

    assert sender_balance >= 0, "Sender balance must be non-negative!!!"

@export
def approve(spender, amount):
    allowed[rt['sender']][spender] = amount


@export
def transfer_from(_from, to, amount):
    assert allowed[_from][rt['sender']] >= amount
    assert balances[_from] >= amount

    allowed[_from][rt['sender']] -= amount
    balances[_from] -= amount
    balances[to] += amount


@export
def allowance(approver, spender):
    return allowed[approver][spender]


@export
def mint(to, amount):
    # print("minting {} to wallet {}".format(amount, to))
    assert rt['sender'] == rt['author'], 'Only the original contract author can mint!'

    balances[to] += amount



from seneca.engine.conflict_resolution import *
from seneca.engine.cr_commands import *
import redis
from unittest import TestCase
import unittest


class TestCRCommandsBase(TestCase):

    def setUp(self):
        self.master = redis.StrictRedis(host='localhost', port=6379, db=0)
        self.working = redis.StrictRedis(host='localhost', port=6379, db=1)

    def tearDown(self):
        self.master.flushdb()
        self.working.flushdb()

    def _new_cr_data(self, sbb_idx=0, finalize=False):
        return CRContext(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, finalize=finalize)

    def _new_cmd(self, sbb_idx=0, contract_idx=0, cr_data=None, finalize=False):
        return CRCmdGetSetBase(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, contract_idx=contract_idx,
                               data=cr_data or self._new_cr_data(sbb_idx=sbb_idx, finalize=finalize))

    def test_add_one_key_to_mod_list(self):
        KEY = 'key_that_was_modified'
        cr_cmd = self._new_cmd()
        cr_cmd.data['getset'].writes[0].add(KEY)

        actual_mods = cr_cmd.data['getset'].writes

        self.assertEqual(len(actual_mods), 1)
        self.assertEqual({KEY}, actual_mods[0])

    def test_add_many_key_to_mod_list(self):
        KEY1 = 'key1_that_was_modified'
        KEY2 = 'key2_that_was_modified'
        cr_cmd = self._new_cmd()
        cr_cmd.data['getset'].writes[0].add(KEY1)
        cr_cmd.data['getset'].writes[0].add(KEY2)

        actual_mods = cr_cmd.data['getset'].writes
        expected_mods = {0: {KEY1, KEY2}}

        self.assertEqual(len(actual_mods), 1)
        self.assertEqual(len(actual_mods[0]), 2)
        self.assertEqual(actual_mods, expected_mods)

    def test_same_keys_to_mod_list(self):
        KEY1 = 'key1_that_was_modified'
        KEY2 = 'key1_that_was_modified'
        cr_cmd = self._new_cmd()
        cr_cmd.data['getset'].writes[0].add(KEY1)
        cr_cmd.data['getset'].writes[0].add(KEY2)

        actual_mods = cr_cmd.data['getset'].writes
        expected_mods = {0: {KEY1}}

        self.assertEqual(len(actual_mods), 1)
        self.assertEqual(len(actual_mods[0]), 1)
        self.assertEqual(actual_mods, expected_mods)

    def test_add_keys_from_different_contract(self):
        KEY1 = 'key1_that_was_modified'
        KEY2 = 'key2_that_was_modified'
        cr_cmd1 = self._new_cmd()
        cr_cmd2 = self._new_cmd(contract_idx=1, cr_data=cr_cmd1.data)
        cr_cmd1.data['getset'].writes[0].add(KEY1)
        cr_cmd2.data['getset'].writes[1].add(KEY2)

        # These guys should share the same modification list (since they are from the same sbb idx)
        self.assertEqual(cr_cmd1.data['getset'].writes, cr_cmd2.data['getset'].writes)
        actual_mods = cr_cmd1.data['getset'].writes

        self.assertEqual(len(actual_mods), 2)
        self.assertEqual(len(actual_mods[0]), 1)
        self.assertEqual(len(actual_mods[1]), 1)
        self.assertEqual(actual_mods[0], {KEY1})
        self.assertEqual(actual_mods[1], {KEY2})

    def test_get_mods_for_sbb_idx_many_keys(self):
        KEY1 = 'key1_that_was_modified'
        KEY2 = 'key2_that_was_modified'
        cr_cmd = self._new_cmd()
        cr_cmd.data['getset'].writes[0].add(KEY1)
        cr_cmd.data['getset'].writes[0].add(KEY2)

        actual_mods = cr_cmd.data['getset'].writes
        expected_mods = {0: {KEY1, KEY2}}

        self.assertEqual(actual_mods, expected_mods)

    def test_adds_empty_set_for_contract_with_no_mods(self):
        KEY2 = 'key2_that_was_modified'
        cr_cmd1 = self._new_cmd()
        cr_cmd2 = self._new_cmd(contract_idx=1, cr_data=cr_cmd1.data)
        # cr_cmd2._add_key_to_mod_list(KEY2)
        cr_cmd2.data['getset'].writes[1].add(KEY2)

        # These guys should share the same modification list (since they are from the same sbb idx)
        self.assertEqual(cr_cmd1.data['getset'].writes, cr_cmd2.data['getset'].writes)
        actual_mods = cr_cmd1.data['getset'].writes

        self.assertEqual(len(actual_mods), 1)
        self.assertEqual(len(actual_mods[0]), 0)
        self.assertEqual(len(actual_mods[1]), 1)
        self.assertEqual(actual_mods[0], set())
        self.assertEqual(actual_mods[1], {KEY2})


if __name__ == "__main__":
    unittest.main()



from tests.utils import TestInterface
from seneca.libs.datatypes import hmap

class TestSubmission(TestInterface):
    def test_floats_to_decimals(self):
        code_str = \
'''
h = 22
j = 7
s = 1.1 + 2.2
print(s)
print(h)
print(h / j)
'''

        self.si.execute_code_str(code_str)

    def test_floats_in_datatypes(self):
        code_str = \
'''

'''



from seneca.libs.math import random

random.seed()

cards = [1, 2, 3, 4, 5, 6, 7, 8]

cardinal_values = ['A', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'J', 'Q', 'K']
suits = ['S', 'C', 'H', 'D']

cities = ['Cleveland', 'Detroit', 'Chicago', 'New York', 'San Francisco']

@export
def shuffle_cards(**kwargs):
    random.shuffle(cards)
    return cards

@export
def random_number(k):
    return random.randrange(k)

@export
def random_number_2(k):
    # adjust the random state by calling another random function
    shuffle_cards()
    return random.randrange(k)

@export
def random_bits(k):
    shuffle_cards()
    shuffle_cards()
    shuffle_cards()
    return random.getrandbits(k)

@export
def int_in_range(a, b):
    shuffle_cards()
    shuffle_cards()
    return random.randint(a, b)

@export
def deal_card():
    random.shuffle(cardinal_values)
    random.shuffle(cardinal_values)
    random.shuffle(cardinal_values)

    random.shuffle(suits)
    random.shuffle(suits)
    random.shuffle(suits)

    c = ''
    c += random.choice(cardinal_values)
    c += random.choice(suits)

    return c

@export
def pick_cities(k):
    return random.choices(cities, k)




import redis
from seneca.libs.logger import get_logger
from seneca.engine.conflict_resolution import CRDataGetSet, CRContext, CRDataBase, CR_EXCLUDED_KEYS


class CRCmdMeta(type):
    def __new__(cls, clsname, bases, clsdict):
        clsobj = super().__new__(cls, clsname, bases, clsdict)
        if not hasattr(clsobj, 'registry'):
            clsobj.registry = {}

        if 'COMMAND_NAME' in clsdict:
            cmd_name = clsdict['COMMAND_NAME']
            assert cmd_name not in clsobj.registry, "Command {} already in registry {}".format(cmd_name, clsobj.registry)
            clsobj.registry[cmd_name] = clsobj

        return clsobj


class CRCmdBase(metaclass=CRCmdMeta):
    DATA_NAME = None

    # TODO -- remove the finalize var. We dont need this.
    def __init__(self, working_db, master_db, sbb_idx: int, contract_idx: int, data: CRContext):
        self.log = get_logger("{}[sbb_{}][contract_{}]".format(type(self).__name__, sbb_idx, contract_idx))
        self.data = data
        self.working, self.master = working_db, master_db
        self.sbb_idx, self.contract_idx = sbb_idx, contract_idx

    def set_params(self, working_db, master_db, sbb_idx: int, contract_idx: int, data: CRContext):
        self.data = data
        self.working, self.master = working_db, master_db
        self.sbb_idx, self.contract_idx = sbb_idx, contract_idx

    def _copy_og_key_if_not_exists(self, key, *args, **kwargs):
        """
        Copies the key from either master db or common layer (working db) to the sub-block specific layer, if it does
        not exist already
        """
        # If the key already exists, bounce out of this method immediately
        if self._sbb_original_exists(key, *args, **kwargs):
            self.log.spam("Key <{}> already exists in sub-block specific data, thus not recopying".format(key))
            return

        # First check the common layer for the key
        if self._db_original_exists(self.working, key, *args, **kwargs):
            self.log.spam("Copying common key <{}> to sb specific data" .format(key))
            self._copy_key_to_sbb_data(self.working, key, *args, **kwargs)

        # Next, check the Master layer for the key
        elif self._db_original_exists(self.master, key, *args, **kwargs):
            self.log.spam("Copying master key <{}> to sb specific data" .format(key))
            self._copy_key_to_sbb_data(self.master, key, *args, **kwargs)

        # Otherwise, if key not found in common or master layer, mark the original as None
        else:
            self._copy_key_to_sbb_data(None, key, *args, **kwargs)

    def _add_key_to_redo_log(self, key, *args, **kwargs):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    def _db_original_exists(self, db, key, *args, **kwargs) -> bool:
        """
        Returns True if 'key' exists on db. False otherwise. args/kwargs can be supplied for more complex
        implementations by subclasses
        :param db: The DB to check
        :param key: The key to check on 'db'
        """
        raise NotImplementedError()

    def _sbb_original_exists(self, key, *args, **kwargs) -> bool:
        """
        Return True if key exists in the sub-block specific data, and False otherwise.
        """
        raise NotImplementedError()

    def _copy_key_to_sbb_data(self, db, key, *args, **kwargs):
        """
        Copies 'key' from the specified to the sub-block specific data
        :param db: The DB to copy the key from. If None, it is implied that the key does not exist in common/master, and
        thus the original value will be set as None
        :param key: The name of the key to copy
        """
        raise NotImplementedError()


class CRCmdExists(CRCmdBase):
    COMMAND_NAME = 'exists'

    def __call__(self, key):
        # TODO this could be made more modular. Current implementation will not scale well --davis
        # First check if key exists in getset
        if key in self.data['getset']:
            return True
        # Next check if key is in HMap
        if key in self.data['hm']:
            return True
        # Then check if it exists in the common layer...
        if self.working.exists(key):
            return True
        # Then finally, check if it exists in the master layer
        return self.master.exists(key)


class CRCmdGetSetBase(CRCmdBase):
    DATA_NAME = 'getset'

    def _db_original_exists(self, db, key) -> bool:
        return db.exists(key)

    def _sbb_original_exists(self, key) -> bool:
        return key in self.data['getset']

    def _copy_key_to_sbb_data(self, db, key):
        val = db.get(key) if db else None
        self.data['getset'][key] = {'og': val, 'mod': None, 'contracts': set()}

    def _get(self, key, return_none=True):
        self._copy_og_key_if_not_exists(key)

        # TODO make all this DRYer so you can abstract it like a pro

        # First, try and return the local modified key
        val = self.data['getset'][key]['mod']
        if val is not None:
            self.log.spam("SBB specific MODIFIED key found for key named <{}>".format(key))
        # Otherwise, default to the local original key
        else:
            self.log.spam("SBB specific ORIGINAL key found for key named <{}>".format(key))
            val = self.data['getset'][key]['og']

        if val is None and not return_none:
            raise Exception("Key '{}' does not exist, but was attempted to be GET".format(key))

        # TODO properly handle CR on stamps key
        if key not in CR_EXCLUDED_KEYS:
            self.data['getset'].reads[self.contract_idx].add(key)
            self.data['getset'][key]['contracts'].add(self.contract_idx)

        return val


# TODO refactor this so they can live in the same base class, and we just specify the command name with a decorator
class CRCmdGet(CRCmdGetSetBase):
    COMMAND_NAME = 'get'

    def __call__(self, key):
        return self._get(key)


class CRCmdSet(CRCmdGetSetBase):
    COMMAND_NAME = 'set'

    def _add_key_to_redo_log(self, key, *args, **kwargs):
        # Return if key already exist in this contract's redo log
        if key in self.data['getset'].redo_log[self.contract_idx]:
            return

        self.data['getset'].redo_log[self.contract_idx][key] = self._get(key, return_none=True)
        self.log.spam("Contract {} added key {} to redo log with val {}".format(self.contract_idx, key,
                                                                                self.data['getset'].redo_log[
                                                                                self.contract_idx][key]))

    def __call__(self, key, value):
        # TODO properly handle CR on stamps key
        if key in CR_EXCLUDED_KEYS:
            return

        assert type(value) in (str, bytes), "Attempted to use 'set' with a value that is not str or bytes (val={}). " \
                                            "This is not supported currently.".format(value)
        self._copy_og_key_if_not_exists(key)
        if type(value) is str:
            value = value.encode()

        self._add_key_to_redo_log(key)

        self.log.spam("Setting SBB specific key <{}> to value {}".format(key, value))
        self.data['getset'][key]['mod'] = value
        self.data['getset'][key]['contracts'].add(self.contract_idx)
        self.data['getset'].writes[self.contract_idx].add(key)
        self.data['getset'].outputs[self.contract_idx] += 'SET {} {};'.format(key, value.decode())


class CRCmdHMapBase(CRCmdBase):
    # MOD_DELIM is used to delimeter the 'key' and 'field' for storing key-fields in the modifications list
    # (key is the name of the hash table, and field is the field on that hash table)
    # TODO this sketches me out. Can people name their keys in such a way that they can do 'sql injection like' attacks?  -- davis
    MOD_DELIM = '*-*'
    DATA_NAME = 'hm'

    def _db_original_exists(self, db, key, field) -> bool:
        return db.hexists(key, field)

    def _sbb_original_exists(self, key, field) -> bool:
        return key in self.data['hm'] and field in self.data['hm'][key]

    def _copy_key_to_sbb_data(self, db, key, field):
        val = db.hget(key, field) if db else None
        self.data['hm'][key][field] = {'og': val, 'mod': None}

    def _get_key_field_name(self, key, field):
        return key + self.MOD_DELIM + field


class CRCmdHGet(CRCmdHMapBase):
    COMMAND_NAME = 'hget'

    def __call__(self, key, field):
        self._copy_og_key_if_not_exists(key, field)

        # TODO make all this DRYer so you can abstract

        # First, try and return the local modified key
        mod_val = self.data['hm'][key][field]['mod']
        if mod_val is not None:
            self.log.spam("SBB specific MODIFIED key found for key named <{}>".format(key))
            return mod_val

        # Otherwise, default to the local original key
        self.log.spam("SBB specific ORIGINAL key found for key named <{}>".format(key))
        return self.data['hm'][key][field]['og']
        # TODO add the read list


class CRCmdHSet(CRCmdHMapBase):
    COMMAND_NAME = 'hset'

    def __call__(self, key, field, value):
        assert type(value) in (str, bytes), "Attempted to use 'hset' with a value that is not str or bytes (val={}). " \
                                            "This is not supported currently.".format(value)
        self._copy_og_key_if_not_exists(key, field)
        if type(value) is str:
            value = value.encode()

        self.log.spam("Setting SBB specific key <{}> to value {}".format(key, value))
        self.data['hm'][key][field]['mod'] = value

        # self._add_key_to_mod_list(self._get_key_field_name(key, field))
        # TODO add the write list



import types
from seneca.engine.interpreter import Seneca


def import_contract(n):
    code = Seneca.interface.get_code_obj(n)
    m = types.ModuleType(n)

    scope = {
        'rt': {
            'contract': m.__dict__['__name__']
        }
    }
    m.__dict__.update(scope)
    Seneca.interface.execute(code, m.__dict__, is_main=False)

    return m



# from unittest import TestCase
# from seneca.engine.interpreter import SenecaInterpreter
# from seneca.engine.module import SenecaFinder, RedisFinder
# import redis, unittest, sys
# from seneca.engine.util import make_n_tup
#
# DO_THING_CODE_STR = """ \
#
# from seneca.contracts.sample import do_that_thing
# print(do_that_thing())
# """
#
# XFER_CODE_STR = """ \
#
# from seneca.contracts.currency import transfer
# transfer('stu', 3)
# """
#
# MINT_CODE_STR = """ \
#
# from seneca.contracts.currency import mint
# mint('davis', 100000)
# mint('stu', 69)
# """
#
#
# class TestInterpreter(TestCase):
#
#     CONTRACTS_TO_STORE = {'runtime_test': 'runtime_test.sen.py', 'sample': 'sample.sen.py',
#                           'currency': 'kv_currency.sen.py'}
#
#     @classmethod
#     def setUpClass(cls):
#         sys.meta_path = [sys.meta_path[2], SenecaFinder(), RedisFinder()]
#         SenecaInterpreter.setup(concurrent_mode=False)
#         SenecaInterpreter.r.flushdb()
#
#         # Store all smart contracts in CONTRACTS_TO_STORE
#         import seneca
#         test_contracts_path = seneca.__path__[0] + '/../test_contracts/'
#
#         for contract_name, file_name in cls.CONTRACTS_TO_STORE.items():
#             with open(test_contracts_path + file_name) as f:
#                 code_str = f.read()
#                 assert not SenecaInterpreter.r.hexists('contracts', contract_name), 'Contract "{}" already exists!'.format(contract_name)
#                 tree, prevalidated = SenecaInterpreter.parse_ast(code_str)
#                 prevalidated_obj = compile(prevalidated, filename='__main__', mode="exec")
#                 SenecaInterpreter.execute(prevalidated_obj)
#                 code_obj = compile(tree, filename='__main__', mode="exec")
#                 SenecaInterpreter.set_code(fullname=contract_name, author='davis', code_obj=code_obj, code_str=code_str)
#
#         cls._mint()
#
#     @classmethod
#     def _exec_code(cls, code_str: str, sender='davis', sbb_idx=0, contract_idx=0, author='davis', master_db_idx=0,
#                    working_db_idx=1):
#         master_db = redis.StrictRedis(host='localhost', port=6379, db=master_db_idx)
#         working_db = redis.StrictRedis(host='localhost', port=6379, db=working_db_idx)
#         SenecaInterpreter.execute(code_str, {
#             'rt': make_n_tup({
#                 'author': author,
#                 'sender': sender
#             }),
#             'sbb_idx': sbb_idx,
#             'contract_idx': contract_idx,
#             'master_db': master_db,
#             'working_db': working_db,
#         })
#
#     @classmethod
#     def _mint(cls):
#         cls._exec_code(MINT_CODE_STR)
#
#     @classmethod
#     def tearDownClass(cls):
#         SenecaInterpreter.r.flushdb()
#         SenecaInterpreter.teardown()
#
#     def test_transfer_with_bookkeeping(self):
#         self._exec_code(XFER_CODE_STR)
#
#
# if __name__ == '__main__':
#     unittest.main()



from unittest import TestCase
from seneca.tooling import *
from decimal import *
GENESIS_AUTHOR = 'anonymoose'
example_code = '''
from seneca.libs.datatypes import hmap

floats = hmap('floats', str, float)

@export
def store_float(s, f):
    floats[s] = f

@export
def read_float(s):
    return floats[s]

@export
def divide_float(s):
    return floats[s] / 2

@export
def add_floats(s1, s2):
    return floats[s1] + floats[s2]
'''


class TestSenecaTooling(TestCase):
    CONTRACTS_TO_STORE = {'currency': 'currency.sen.py'}

    def setUp(self):
        self.driver = SenecaInterface(concurrent_mode=False,
                           port=6379,
                           password='')
        self.driver.r.flushdb()
        import seneca
        test_contracts_path = seneca.__path__[0] + '/../test_contracts/'

        for contract_name, file_name in self.CONTRACTS_TO_STORE.items():
            with open(test_contracts_path + file_name) as f:
                code_str = f.read()
                self.driver.publish_code_str(contract_name, GENESIS_AUTHOR, code_str)

    def test_submission(self):
        self.driver.publish_code_str(
            'example_code',
            author='stuart',
            code_str=example_code,
        )

        self.assertEqual(example_code, self.driver.get_contract_meta('example_code')['code_str'])

    def test_wrapper(self):
        self.driver.publish_code_str(
            'example_code',
            author='stuart',
            code_str=example_code,
        )

        wrapper = ContractWrapper('example_code', driver=self.driver, default_sender='stuart')
        expected_functions = {'store_float', 'read_float', 'divide_float', 'add_floats'}

        self.assertEqual(expected_functions - set(dir(wrapper)), set())

    def test_call_function(self):
        self.driver.publish_code_str(
            'example_code',
            author='stuart',
            code_str=example_code,
        )

        wrapper = ContractWrapper('example_code', driver=self.driver, default_sender='stuart')
        wrapper.store_float(s='stu', f=0.1234)
        f = wrapper.read_float(s='stu')

        self.assertEqual(f['output'], Decimal('0.1234'))

if __name__ == '__main__':
    import unittest
    unittest.main()



"""Module for initializing settings related to the built-in seneca logger
Functions:
-get_logger"""

import logging, coloredlogs
import os, sys

VALID_LVLS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']
_LOG_LVL = os.getenv('LOG_LEVEL', None)
if _LOG_LVL:
    assert _LOG_LVL in VALID_LVLS, "Log level {} not in valid levels {}".format(_LOG_LVL, VALID_LVLS)
    _LOG_LVL = getattr(logging, _LOG_LVL)
else:
    _LOG_LVL = 1

req_log = logging.getLogger('urllib3')
req_log.setLevel(logging.WARNING)
req_log.propagate = True

def get_main_log_path():
    from seneca.libs import logger

    root = logger.__file__  # resolves to '/Users/davishaba/Developer/seneca/seneca/logger/__init__.py'
    log_path = '/'.join(root.split('/')[:-3]) + '/logs/seneca.log'

    # Create log directory if it does not exist
    log_dir = os.path.dirname(log_path)
    try:
        if not os.path.isdir(log_dir):
            os.makedirs(log_dir)
    except Exception as e:
        print("Possible error creating log file: {}".format(e))

    return log_path


format = '%(asctime)s.%(msecs)03d %(name)s[%(process)d][%(processName)s] <{}> %(levelname)-2s %(message)s'.format(os.getenv('HOST_NAME', 'Node'))

"""
Custom Log Levels
"""

CUSTOM_LEVELS = {
    'SPAM': 1,
    'DEBUGV': 5,
    'SOCKET': 23,
    'NOTICE': 24,
    'SUCCESS': 26,
    'SUCCESS2': 27,
    'IMPORTANT': 56,
    'IMPORTANT2': 57,
    'IMPORTANT3': 58,
    'FATAL': 9001,
    'TEST': 9002,
    }

for log_name, log_level in CUSTOM_LEVELS.items():
    logging.addLevelName(log_level, log_name)


def apply_custom_level(log, name: str, level: int):
    def _lvl_func(message, *args, **kws):
        if level >= log.getEffectiveLevel():
            log._log(level, message, args, **kws)

    setattr(log, name.lower(), _lvl_func)


"""
Custom Styling
"""

coloredlogs.DEFAULT_LEVEL_STYLES = {
    'critical': {'color': 'white', 'bold': True, 'background': 'red'},
    'fatal': {'color': 'white', 'bold': True, 'background': 'red', 'underline': True},
    'debug': {'color': 'green'},
    'error': {'color': 'red'},
    'info': {'color': 'white'},
    'notice': {'color': 'magenta'},
    'test': {'color': 'magenta'},
    'socket': {'color': 216},
    'important': {'color': 'cyan', 'bold': True, 'background': 'magenta'},
    'important2': {'color': 'magenta', 'bold': True, 'background': 'cyan'},
    'important3': {'color': 'black', 'bold': True, 'background': 'yellow'},
    'spam': {'color': 'white', 'faint': True},
    'success': {'color': 'white', 'bold': True, 'background': 'green'},
    'success2': {'color': 165, 'bold': True, 'background': 'green'},
    'verbose': {'color': 'blue'},
    'warning': {'color': 'yellow'},
    'debugv': {'color': 'blue', 'faint': False}
}
coloredlogs.DEFAULT_FIELD_STYLES = {
    'asctime': {'color': 'green'},
    'hostname': {'color': 'magenta'},
    'levelname': {'color': 'black', 'bright': True},
    'name': {'color': 'blue'},
    'programname': {'color': 'cyan'}
}


class LoggerWriter:
    def __init__(self, level):
        self.level = level
    def write(self, message):
        if message != '\n':
            self.level(message)
    def flush(self):
        return


class ColoredFileHandler(logging.FileHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFormatter(
            coloredlogs.ColoredFormatter(format)
        )


class ColoredStreamHandler(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFormatter(
            coloredlogs.ColoredFormatter(format)
        )


def _ignore(*args, **kwargs):
    pass

class MockLogger:
    def __getattr__(self, item):
        return _ignore


def get_logger(name=''):
    if _LOG_LVL < 0:
        return MockLogger()

    filedir = "logs/{}".format(os.getenv('TEST_NAME', 'test'))
    filename = "{}/{}.log".format(filedir, os.getenv('HOST_NAME', name))

    try:
        if not os.path.isdir(filedir):
            os.makedirs(filedir, exist_ok=True)
    except Exception as e:
        print("Possible error creating log file: {}".format(e))

    filehandlers = [
        logging.FileHandler(get_main_log_path(), delay=True),
        logging.FileHandler(filename, delay=True),
        ColoredFileHandler('{}_color'.format(filename), delay=True),
        ColoredStreamHandler()
    ]
    logging.basicConfig(
        format=format,
        handlers=filehandlers,
        level=logging.DEBUG
    )

    log = logging.getLogger(name)
    log.setLevel(_LOG_LVL)

    if os.getenv('HOST_IP'):
        sys.stdout = LoggerWriter(log.debug)
        sys.stderr = LoggerWriter(log.error)

    for log_name, log_level in CUSTOM_LEVELS.items():
        apply_custom_level(log, log_name, log_level)

    return log


def overwrite_logger_level(level):
    global _LOG_LVL
    _LOG_LVL = level

    for name in logging.Logger.manager.loggerDict.keys():
        log = logging.getLogger(name)
        log.setLevel(level)



import random, seneca
from os import getenv as env
from dotenv import load_dotenv


def load_env():
    load_dotenv(dotenv_path='{}/../docker/redis.env'.format(path), override=True)


def get_redis_port(port=None):
    if port is not None:
        return port

    if env('CIRCLECI'):
        return 6379

    return env('REDIS_PORT', 6379)


def get_redis_password(password=None):
    if password is not None:
        return password

    if env('CIRCLECI'):
        return ''

    return env('REDIS_PASSWORD', '')

path = seneca.__path__[0]
load_env()

MASTER_DB = 0
DB_OFFSET = 1
CODE_OBJ_MAX_CACHE = 64

# Number of available db's SenecaClients have available to get ahead on the next sub block while other sb's are
# awaiting a merge confirmation
NUM_CACHES = 2

# Number of sb's to queue up if we run out of caches
MAX_SB_QUEUE_SIZE = 8



from seneca.libs.datatypes import hmap
from seneca.contracts.tau import transfer, get_balance

owners = hmap('owners', str, str)
prices = hmap('prices', str, int)
colors = hmap('colors', str, str)

max_x = 250
max_y = 250

@export
def coor_str(x, y):
    return '{},{}'.format(x, y)

def owner_of_pixel(x, y):
    return owners[coor_str(x, y)]

def price_of_pixel(x, y):
    return prices[coor_str(x, y)]

@export
def buy_pixel(x, y, r, g, b, new_price=0):
    assert new_price >= 0

    x, y, r, g, b = [int(i) for i in [x, y, r, g, b]]

    assert x >= 0 and x < max_x
    assert y >= 0 and y < max_y
    assert r >= 0 and r < 256
    assert g >= 0 and g < 256
    assert b >= 0 and b < 256

    price = price_of_pixel(x, y)

    assert get_balance(rt['sender']) >= price

    owner = owner_of_pixel(x, y)

    transfer(owner, price)

    owners[coor_str(x, y)] = rt['sender']
    prices[coor_str(x, y)] = new_price if new_price > 0 else price
    colors[coor_str(x, y)] = '{},{},{}'.format(r, g, b)


@export
def price_pixel(x, y, amount):
    owner = owner_of_pixel(x, y)
    assert rt['sender'] == owner

    x = int(x)
    y = int(y)

    amount = int(amount)
    prices[coor_str(x, y)] = amount

@export
def color_pixel(x, y, r, g, b):
    owner = owner_of_pixel(x, y)
    assert rt['sender'] == owner

    r = int(r)
    g = int(g)
    b = int(b)

    colors[coor_str(x, y)] = '{},{},{}'.format(r, g, b)



import redis, ast, marshal, array, copy, inspect, types, uuid, copy, ujson as json, sys, time
from seneca.constants.whitelists import ALLOWED_AST_TYPES, ALLOWED_IMPORT_PATHS, SAFE_BUILTINS, SENECA_LIBRARY_PATH
from seneca.constants.config import get_redis_port, get_redis_password, MASTER_DB, DB_OFFSET, CODE_OBJ_MAX_CACHE
from functools import lru_cache
from seneca.libs.metering.tracer import Tracer
import seneca, os
from os.path import join
from seneca.engine.book_keeper import BookKeeper

class ReadOnlyException(Exception):
    pass


class CompilationException(Exception):
    pass


class Seneca:
    current_ast_types = None
    prevalidated = None
    postvalidated = None
    protected_variables = None
    concurrent_mode = True
    interface = None

    CURRENCY_NAME = 'currency'

    exports = {}
    imports = {}
    loaded = {}
    cache = {}
    resources = {}
    methods = {}

    basic_scope = {}


class ScopeParser:
    def set_scope(self, fn, args, kwargs):
        fn.__globals__.update(Seneca.loaded['__main__'])

        if fn.__globals__.get('__use_locals__') == '{}.{}'.format(fn.__module__.rsplit('.')[-1], fn.__name__):
            if fn.__globals__.get('__args__'): args = fn.__globals__['__args__']
            if fn.__globals__.get('__kwargs__'): kwargs = fn.__globals__['__kwargs__']
        else:
            fn.__globals__['rt']['contract'] = fn.__module__

        return args, kwargs

    def set_scope_during_compilation(self, fn):
        self.module = '.'.join([fn.__module__, fn.__name__])
        Seneca.exports[self.module] = True

class Function(ScopeParser):
    def __call__(self, fn):
        def _fn(*args, **kwargs):
            args, kwargs = self.set_scope(fn, args, kwargs)
            BookKeeper.set_info(rt=fn.__globals__['rt'])
            return fn(*args, **kwargs)
        return _fn

class Export(ScopeParser):
    def __call__(self, fn):
        if not fn.__module__: return
        self.set_scope_during_compilation(fn)
        def _fn(*args, **kwargs):
            args, kwargs = self.set_scope(fn, args, kwargs)
            BookKeeper.set_info(rt=fn.__globals__['rt'])
            return fn(*args, **kwargs)
        return _fn


class Seed(ScopeParser):
    def __call__(self, fn):
        if fn.__globals__.get('__seed__') == True:
            old_concurrent_mode = Seneca.concurrent_mode
            Seneca.concurrent_mode = False
            BookKeeper.set_info(rt=fn.__globals__['rt'])
            fn()
            Seneca.concurrent_mode = old_concurrent_mode


Seneca.basic_scope = {
    'export': Export(),
    'seed': Seed(),
    '__function__': Function(),
    '__builtins__': SAFE_BUILTINS,
    '__use_locals__': False
}


class SenecaNodeTransformer(ast.NodeTransformer):
    def generic_visit(self, node):
        Seneca.current_ast_types.add(type(node))
        return super().generic_visit(node)

    def visit_Name(self, node):
        if SenecaInterpreter.is_system_variable(node.id):
            raise CompilationException('Not allowed to read "{}"'.format(node.id))
        self.generic_visit(node)
        return node

    def visit_Attribute(self, node):
        if SenecaInterpreter.is_system_variable(node.attr):
            raise CompilationException('Not allowed to read "{}"'.format(node.attr))
        self.generic_visit(node)
        return node

    def visit_Import(self, node):
        module_name = node.names[0].name
        SenecaInterpreter.assert_import_path(module_name)
        Seneca.prevalidated.body.append(node)
        self.generic_visit(node)
        Seneca.postvalidated.body.append(node)
        return node

    def visit_ImportFrom(self, node):
        module_name = node.names[0].name
        SenecaInterpreter.assert_import_path(node.module, module_name=module_name)
        Seneca.prevalidated.body.append(node)
        self.generic_visit(node)
        Seneca.postvalidated.body.append(node)
        return node

    def visit_Assign(self, node):

        for target in node.targets:
            if type(target) == ast.Tuple:
                items = []
                for item in target.elts:
                    SenecaInterpreter.check_protected(item, Seneca.protected_variables)
                    items.append(item.id)
                val = None
                if type(node.value) == ast.Call:
                    val = node.value.func.id
                elif type(target) == ast.Tuple:
                    val = 'tuple'
                Seneca.resources[', '.join(items)] = val
            elif type(node.value) == ast.Call:
                if hasattr(node.value.func, 'id'):
                    Seneca.resources[target.id] = node.value.func.id
                SenecaInterpreter.check_protected(target, Seneca.protected_variables)
            else:
                Seneca.resources[target.id] = node.value.__class__.__name__
                SenecaInterpreter.check_protected(target, Seneca.protected_variables)

        if type(node.value) == ast.Call:
            Seneca.postvalidated.body.append(node)
        self.generic_visit(node)
        return node

    def visit_AugAssign(self, node):
        SenecaInterpreter.check_protected(node.target, Seneca.protected_variables)
        self.generic_visit(node)
        return node

    def visit_Num(self, node):
        if isinstance(node.n, float) or isinstance(node.n, int):
            return ast.Call(func=ast.Name(id='make_decimal', ctx=ast.Load()),
                            args=[node], keywords=[])
        self.generic_visit(node)
        return node

    def visit_FunctionDef(self, node):
        for item in node.body:
            if type(item) in [ast.ImportFrom, ast.Import]:
                raise CompilationException('Not allowed to import inside a function definition')
        set_scope = True
        for item in node.decorator_list:
            if item.id == 'export':
                Seneca.exports[node.name] = True
                Seneca.methods[node.name] = [arg.arg for arg in node.args.args]
                set_scope = False
            elif item.id == 'seed':
                set_scope = False
        if set_scope:
            node.decorator_list.append(
                ast.Name(id='__function__', ctx=ast.Load())
            )
        return node


class SenecaInterpreter:
    def __init__(self, concurrent_mode=True, port=None, password=None, bypass_currency=False):
        self.r = redis.StrictRedis(host='localhost',
                                  port=get_redis_port(port=port),
                                  db=MASTER_DB,
                                  password=get_redis_password(password=password)
                                  )

        self.bypass_currency = bypass_currency
        self.setup_tracer()
        Seneca.concurrent_mode = concurrent_mode

    def setup_tracer(self):
        seneca_path = seneca.__path__[0]
        path = join(seneca_path, 'constants', 'cu_costs.const')
        os.environ['CU_COST_FNAME'] = path
        self.tracer = Tracer()

    def get_code_obj(self, fullname):
        code_obj = self.r.hget('contracts', fullname)
        assert code_obj, 'User module "{}" not found!'.format(fullname)
        return marshal.loads(code_obj)

    def get_code_str(self, fullname):
        meta = self.get_contract_meta(fullname)
        assert meta.get('code_str'), 'Cannot find original code string for module "{}" not found!'.format(fullname)
        return meta['code_str']

    def get_contract_meta(self, fullname):
        byte_str = self.r.hget('contracts_meta', fullname)
        meta = json.loads(byte_str)
        return meta

    def set_code(self, fullname, tree_obj, code_obj, code_str, author):
        pipe = self.r.pipeline()
        pipe.hset('contracts', fullname, marshal.dumps(tree_obj))
        pipe.hset('contracts_code', fullname, marshal.dumps(code_obj))
        pipe.hset('contracts_meta', fullname, json.dumps({
            'code_str': code_str,
            'resources': Seneca.resources,
            'methods': Seneca.methods,
            'author': author,
            'timestamp': time.time()
        }))
        pipe.execute()

    def remove_code(self, fullname):
        pipe = self.r.pipeline()
        pipe.hdel('contracts', fullname)
        pipe.hdel('contracts_code', fullname)
        pipe.hdel('contracts_meta', fullname)
        pipe.execute()

    @staticmethod
    def parse_ast(code_str, protected_variables=[]):

        # crude but effective way to force fixed precision without developer behavior change
        decimal_addon = 'from seneca.libs.decimal import make_decimal'

        code_str = decimal_addon + '\n' + code_str

        Seneca.current_ast_types = set()
        Seneca.protected_variables = protected_variables

        tree = ast.parse(code_str)
        Seneca.resources = {}
        Seneca.methods = {}
        Seneca.protected_variables += ['export']
        Seneca.prevalidated = copy.deepcopy(tree)
        Seneca.prevalidated.body = []
        Seneca.postvalidated = copy.deepcopy(tree)
        Seneca.postvalidated.body = []

        tree = SenecaNodeTransformer().visit(tree)
        ast.fix_missing_locations(tree)

        illegal_ast_nodes = Seneca.current_ast_types - ALLOWED_AST_TYPES
        assert not illegal_ast_nodes, 'Illegal AST node(s) in module: {}'.format(
            ', '.join(map(str, illegal_ast_nodes)))

        return tree, Seneca.postvalidated, Seneca.prevalidated

    @staticmethod
    def validate():
        for import_path in Seneca.imports:
            if not Seneca.exports.get(import_path):
                raise CompilationException('Forbidden to import "{}"'.format(
                    import_path))

    @staticmethod
    def is_system_variable(v):
        return v.startswith('__') and v.endswith('__')

    @staticmethod
    def check_protected(target, protected_variables):
        if isinstance(target, ast.Subscript):
            return
        if target.id in protected_variables \
            or target.id in [k.rsplit('.', 1)[-1] for k in Seneca.imports.keys()]:
            raise ReadOnlyException('Cannot assign value to "{}" as it is a read-only variable'.format(target.id))

    @staticmethod
    def assert_import_path(import_path, module_name=None):
        if module_name == '*':
            raise ImportError('Not allowed to import *')
        elif module_name:
            import_path = '.'.join([import_path, module_name])
        if import_path.startswith(SENECA_LIBRARY_PATH):
            Seneca.exports[import_path] = True
            return True
        for path in ALLOWED_IMPORT_PATHS:
            if import_path.startswith(path):
                if len(import_path.split('.')) - len(path.split('.')) == 2:
                    if not Seneca.exports.get(import_path):
                        Seneca.imports[import_path] = True
                    return True
                else:
                    raise ImportError('Instead of importing the entire "{}" module, you must import each functions directly.'.format(import_path))
        raise ImportError('Cannot find module "{}" in allowed protected_imports'.format(import_path))

    def execute(self, code, scope={}, is_main=True):
        scope.update(Seneca.basic_scope)
        if is_main:
            Seneca.loaded['__main__'] = scope
        Seneca.loaded['__main__'].update(scope)

        exec(code, scope)
        if is_main:
            self.validate()

    @lru_cache(maxsize=CODE_OBJ_MAX_CACHE)
    def get_cached_code_obj(self, module_path, stamps_supplied):
        self.assert_import_path(module_path)
        module = module_path.rsplit('.', 1)
        meta = self.get_contract_meta(module[0].rsplit('.')[-1])
        if stamps_supplied != None:
            import_obj = compile('''
from seneca.contracts.currency import submit_stamps
submit_stamps({})
from {} import {}
            '''.format(stamps_supplied, module[0], module[1]), '__main__', 'exec')
        else:
            import_obj = compile('''
from {} import {}
            '''.format(module[0], module[1]), '__main__', 'exec')
        code_str = '''
result = {}()
        '''.format(module[1])
        fn_call_obj = compile(code_str, '__main__', 'exec')

        return fn_call_obj, import_obj, meta

    def execute_function(self, module_path, sender, stamps, *args, **kwargs):
        module_name = module_path.rsplit('.', 1)[0]
        contract_name = module_name.rsplit('.', 1)[-1]
        fn_call_obj, import_obj, meta = self.get_cached_code_obj(module_path, stamps)
        scope = {
            'rt': {'author': meta['author'], 'sender': sender, 'contract': contract_name},
            '__args__': args,
            '__kwargs__': kwargs,
        }

        currency_module_path = module_path.rsplit('.', 1)[0] + '.' + Seneca.CURRENCY_NAME
        c_fn_call_obj, c_import_obj, c_meta = self.get_cached_code_obj(module_path, stamps)
        currency_scope = {
            'rt': {'author': c_meta['author'], 'sender': sender, 'contract': Seneca.CURRENCY_NAME},
        }

        scope.update(Seneca.basic_scope)

        if not self.bypass_currency:
            currency_scope.update(Seneca.basic_scope)

            Seneca.loaded['__main__'] = currency_scope
            _obj = marshal.loads(self.r.hget('contracts_code', Seneca.CURRENCY_NAME))
            exec(_obj, currency_scope)  # rebuilds RObjects for currency contract

        Seneca.loaded['__main__'] = scope
        _obj = marshal.loads(self.r.hget('contracts_code', contract_name))
        exec(_obj, scope)  # rebuilds RObjects
        exec(import_obj, scope)  # submits stamps

        scope.update({'__use_locals__': '.'.join(module_path.split('.')[-2:])})
        if stamps is not None:
            self.tracer.set_stamp(stamps)
            self.tracer.start()

            try:
                exec(fn_call_obj, scope)  # Actually execute the function
            except Exception as e:
                raise e
            finally:
                self.tracer.stop()

            stamps -= self.tracer.get_stamp_used()
        else:
            exec(fn_call_obj, scope)
            stamps = 0
        return {
            'status': 'success',
            'output': scope.get('result'),
            'remaining_stamps': stamps
        }



import sys, importlib, warnings
from seneca.engine.module import SenecaFinder, RedisFinder
from seneca.engine.interpreter import SenecaInterpreter, Seneca
import inspect
import autopep8


class SenecaInterface(SenecaInterpreter):
    """
        High level API for interacting with Seneca Smart Contracts
    """

    def __init__(self, *args, **kwargs):
        if not isinstance(sys.meta_path[2], RedisFinder):
            self.old_sys_path = sys.meta_path
            sys.meta_path = [sys.meta_path[2], SenecaFinder(), RedisFinder()]
        Seneca.interface = self
        super().__init__(*args, **kwargs)

    def __enter__(self, *args, **kwargs):
        self.old_concurrent_mode = Seneca.concurrent_mode
        return self

    def __exit__(self, type, value, traceback):
        Seneca.concurrent_mode = self.old_concurrent_mode
        return False

    @staticmethod
    def function_to_code_string(f):
        _code = inspect.getsourcelines(f)[0]
        _code = _code[1:]
        code_str = ''
        for c in _code:
            code_str += c

        standard_indented_code = autopep8.fix_code(code_str, options={'select': ['E101']})

        final_code = ''
        for line in standard_indented_code.split('\n'):
            if line.startswith('    '):
                final_code += line[4:] + '\n'

        final_code = autopep8.fix_code(final_code)

        return final_code

    def compile_code(self, code_str, scope={'rt': {'sender': 'anonymous', 'contract': 'arbitrary'}}):
        tree, code, prevalidated = self.parse_ast(code_str, protected_variables=list(scope.keys()))
        prevalidated_obj = compile(prevalidated, filename='__main__', mode="exec")
        self.execute(prevalidated_obj, scope)
        tree_obj = compile(tree, filename='__main__', mode="exec")
        code_obj = compile(code, filename='__main__', mode="exec")
        scope.update({'__seed__': True})
        self.execute(tree_obj, scope)
        return tree_obj, code_obj

    def execute_code_str(self, code_str, scope={'rt': {'sender': 'anonymous', 'contract': 'arbitrary'}}):
        Seneca.imports = {}
        tree_obj, code_obj = self.compile_code(code_str, scope)
        return self.execute(tree_obj, scope)

    def publish_code_str(self, fullname, author, code_str, scope={}):
        assert not self.r.hexists('contracts', fullname), 'Contract "{}" already exists!'.format(fullname)
        Seneca.imports = {}
        tree_obj, code_obj = self.compile_code(code_str, scope={'rt': {'author': author, 'contract': fullname}})
        self.set_code(fullname, tree_obj, code_obj, code_str, author)

    def publish_function(self, f, contract_name, author, scope={}):
        code_str = self.function_to_code_string(f)
        assert not self.r.hexists('contracts', contract_name), 'Contract "{}" already exists!'.format(contract_name)
        Seneca.imports = {}
        tree_obj, code_obj = self.compile_code(code_str, scope={'rt': {'author': author, 'contract': contract_name}})
        self.set_code(fullname=contract_name, tree_obj=tree_obj, code_obj=code_obj, code_str=code_str, author=author)



from seneca.libs.datatypes import hmap

balances = hmap('balances', str, int)

@seed
def deposit_to_all_wallets():
    balances['birb'] = 1000000

def balance_of(wallet_id):
    return balances[wallet_id]

@export
def transfer(to, amount):
    print("transfering from {} to {} with amount {}".format(rt['sender'], to, amount))
    balances[rt['sender']] -= amount
    balances[to] += amount
    sender_balance = balances[rt['sender']]

    assert sender_balance >= 0, "Sender balance must be non-negative!!!"
